# -*- test-case-name: atop.test.test_credup -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""A collection of classes and utilities related to accessing remote items in a
sensible fashion.
"""

from atop.credup import ICredentialsChecker
from atop.store import Store, SubStore

from twisted.internet import reactor, defer
from twisted.python import failure
from twisted.python.components import Interface, Adapter, registerAdapter

def _callInD(d, f, a, k):
    try:
        r = f(*a,**k)
    except:
        r = failure.Failure()
    d.callback(r)

def deferLater(s, f, *a, **k):
    d = defer.Deferred()
    reactor.callLater(s, _callInD, d, f, a, k)
    return d

class NearReference:
    """
    A faux-far reference to a persistent object that might have been 'far
    away', potentially on another server in the cluster, but is actually near.
    Generally you don't create these directly: call
    IFarIndex(self.store).createRemote() if you want to create an object, call
    IFarIndex(self.store).farByName(name, domain).
    """

    def __init__(self, actuallyNear):
        """Create a far reference.

        @param actuallyNear: Actually, this is a near reference.  Currently the
        only way to specify a far reference at this point.
        """
        self.actuallyNear = actuallyNear

    def far(self, interface, methodName, *a, **kw):
        """This will use the reactor's callLater to make sure that the *logic*
        is actually interleaveable.  It will open a new transaction for this
        particular call.

        This doesn't enforce ANYTHING about the objects being passed, so please
        make sure that they are immutable or unshared non-persistent objects,
        or they are being passed to a method which does not mutate them.
        Passing items to this method will confuse the control flow just as
        badly as manually avoiding using Deferreds.
        """
        return deferLater(0, self.actuallyNear.store.transact,
                          lambda : getattr(interface(self.actuallyNear), methodName)(*a,**kw))

class FailedFar:
    def __init__(self, f=None):
        if f == None:
            f = failure.Failure()
        self.failure = f

    def far(self, interface, methodName, *a, **kw):
        return defer.fail(self.failure)

class IFarIndex(Interface):
    """Marker interface for FarIndex.
    """

class FarIndex(Adapter):
    """Index of which items are possibly stored remotely.
    """
    temporaryAdapter = True             # it's not even stateful
    __implements__ = IFarIndex
    defaultDomain = 'localhost'

    def farByName(self, localpart, domain=None):
        """I won't manage deferring here.  You just get a reference immediately
        and then make .far() requests of it.  Any of those requests may fail
        with a HostNotAvailable or whatever, though.
        """
        try:
            return NearReference(ICredentialsChecker(self.original).getAvatar(localpart, domain))
        except:
            return FailedFar()

def farIndexForSubStore(s):
    return IFarIndex(s.store, persist=False)


registerAdapter(FarIndex, Store, IFarIndex)

registerAdapter(farIndexForSubStore, SubStore, IFarIndex)
