# -*- test-case-name: atop.test.test_powerup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.python.components import Interface, Adapter, registerAdapter
from twisted.python import reflect

from atop.store import Item, Pool, NameIndex, SubStore, Store

class IPowerup(Interface):

    property("name", doc="""
    The nickname for this powerup.  This may or may not be mutable: if it is a
    system-configured powerup like 'to-do', it needs its name to be consistent
    so that multiple things can find the root todo pool to add items to it.  If
    it is a user-created pool like 'Bob's Favorite Photos From his Vacation in
    Hali', then it is user-defined and therefore mutable.
    """)

    property("commandPrefixes",
             doc="""
    This property is a list of strings which are the commands that this powerup
    will recognize.  For example, for a To-Do system, this might be ['todo',
    't'], which would mean that from the web-UI, the `search box' would
    activate the To-Do system's commandLine method with either `/t' or `/todo'.
    """)

    property("dependencies",
             doc="""
    A list of other implementors of IPowerup which must be installed on the
    avatar before we can consider it ready for this powerup to be finished
    instantiating.
    """)

    property("storeDependencies",
             doc="""
    A list of other implementors of IPowerup which must be installed on the
    avatar's store before we can consider it ready for this powerup to be
    finished instantiating.
    """)

    def commandLine(self, commander, context, text):
        """
        @param commander: The object that initiated this command-line action.

        @param context: The context in which this action was invoked.  In the
        case of a to-do being created from the web, the context might be the
        current resource.

        @param text: the text following the command string and separator.  So
        if the user types `/todo call dentist', this parameter is 'call
        dentist'.
        """


class PowerupBase(Item):
    """Absolute base-est class for powerups.  Use this if you need to execute
    before the PowerStation is installed on the avatar.
    """
    __implements__ = IPowerup,
    dependencies = ()
    storeDependencies = ()
    powerupName = 'Base'

    def __init__(self, avatar):
        self.loadDependencies(avatar)
        if avatar.store:
            self.loadStoreDependencies(avatar.store)
        Item.__init__(self, avatar)
        self.name = self.powerupName
        self.setUpPools(avatar)

    def loadDependencies(self, avatar):
        pm = IPowerStation(avatar)
        L = []
        reflect.accumulateClassList(self.__class__, 'dependencies', L)
        for dependency in L:
            pm.installPowerup(dependency)

    def loadStoreDependencies(self, store):
        pm = IPowerStation(store)
        L = []
        reflect.accumulateClassList(self.__class__, 'storeDependencies', L)
        for dependency in L:
            pm.installPowerup(dependency)

    def setUpPools(self, avatar):
        pass

class IPowerStation(Interface):
    """Tag interface for PowerStation, don't implement me.
    """

class PowerStation(Adapter):
    """Adapter around a store or substore which manages the creation and
    dependency of other powerups.
    """
    def __init__(self, original):
        Adapter.__init__(self, original)
        self.powerups = {}

    def installPowerup(self, powerupClass):
        self.original.touch()
        cn = reflect.qual(powerupClass)
        if self.powerups.has_key(cn):
            return self.powerups[cn].getItem()
        else:
            pup = powerupClass(self.original)
            self.powerups[cn] = pup.referenceTo()
            pups = self.getPowerupsPool()
            if pup not in pups:
                pups.addItem(pup)
            return pup

    def getPowerupsPool(self):
        return self._powerupsPool.getItem()

registerAdapter(PowerStation, SubStore, IPowerStation)

class MetaPowerup(PowerupBase):
    invisible = True
    powerupName = 'MetaPowerup'
    def setUpPools(self, avatar):
        pm = IPowerStation(avatar)
        if isinstance(avatar, Store):
            # short-circuit for now, while the implementation of PowerStation is
            # divergent...
            return
        p = Pool(avatar, "powerups")
        p.invisible = True
        p.addIndex(NameIndex())
        p.suggestIndex('name', False)
        pm._powerupsPool = p.referenceTo()
        avatar.rootPoolRef.getItem().addItem(p)
        self.invisible = True

class Powerup(PowerupBase):
    """
    Base class for normal powerups.
    """
    dependencies = [MetaPowerup]
    powerupName = 'Powerup'


class Storeup(Powerup):
    """I am a store powerup.  I can be set as a component on the root store.
    """
    implementsForStore = None
    powerupName = 'Storage'

    def setindex_powerup(self, newpowerup):
        self.implementsForStore = newpowerup
        return newpowerup

    def index_powerup(self, pool):
        return self.implementsForStore
