# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from os.path import join as opj

from atop.test import unittest
from atop.sched import IScheduler, SystemScheduler, SubScheduler, SchedulerService, SchedulerBase, Schedulable, CallingItem
from atop.powerup import IPowerStation
from atop.store import Store, SubStore, Item

class MySch(Item, Schedulable):
    count = 0
    def run(self):
        self.count += 1

class SchedTest(unittest.TestCase):

    def dbSetUp(self):
        self.times = [0.0, 1.0, 2.0, 3.0, 10.0, 20.0, 30.0]
        self.calls = []
        self.origSchedBase = SchedulerBase.clock
        SchedulerBase.clock = staticmethod(lambda : self.times[0])

        def callLater(t, f, *a, **b):
            self.calls.append((t, f))

        self.origCallLater = SystemScheduler.callLater
        SystemScheduler.callLater = staticmethod(callLater)

        self.svc = SchedulerService(self.store)
        self.sub = SubStore(self.store)
        IPowerStation(self.sub).installPowerup(SubScheduler)
        self.svc.startService()

    def dbTearDown(self):
        SchedulerBase.clock = self.origSchedBase
        SystemScheduler.callLater = self.origCallLater

    def dbtestTopScheduler(self):
        msc = MySch(self.store)
        msc.schedule(2)
        for x in range(10):
            t, f = self.calls.pop(0)
            f()
        self.assertEquals(msc.count, 0)
        for x in range(3):
            self.times.pop(0)
            t, f = self.calls.pop(0)
            f()
        self.assertEquals(msc.count, 1)
        self.assertEquals(self.calls, [])

    def dbtestSubScheduler(self):
        msc = MySch(self.sub)
        msc.schedule(2.5)
        for x in range(10):
            t, f = self.calls.pop(0)
            f()
        self.assertEquals(msc.count, 0)
        for x in range(3):
            self.times.pop(0)
            t, f = self.calls.pop(0)
            f()
        self.assertEquals(msc.count, 1)
        self.assertEquals(self.calls, [])

        self.failIf(list(self.svc.sched.pool.queryIndex()), "Items left in the scheduled pool")

def testCallingItem(xRef, y):
    x = xRef.getItem()
    x.touch()
    x.y = y

class SchedulerStub:
    def __init__(self, times, calls):
        self.times = times
        self.calls = calls

    def clock(self):
        return self.times[0]

    def callLater(self, n, f, *a, **kw):
        self.calls.append((n, f, a, kw))

class CallingItemTestCase(unittest.TestCase):

    storePowerups = [SystemScheduler]

    def dbSetUp(self):
        self.times = [0, 1, 2, 3, 4, 5]
        self.delayedCalls = []
        self.stub = SchedulerStub(self.times, self.delayedCalls)
        self.sched = IScheduler(self.store)
        self.sched.clock = self.stub.clock
        self.sched.callLater = self.stub.callLater

    def dbtestCalling(self):
        x = Item(self.store)
        c = CallingItem(self.store, 3, testCallingItem, x.referenceTo(), 'foo')
        for i in range(4):
            self.times.pop(0)
            n, f, a, kw = self.delayedCalls.pop(0)
            f(*a, **kw)
        self.assertEquals(x.y, 'foo')
        self.assertRaises(KeyError, self.store.getItemByID, c.storeID)
