
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from twisted.trial import unittest
from twisted.internet import protocol
from twisted.python import components

from atop import tpython

class IFakeTransport(components.Interface):
    pass

class FakeTransport:
    __implements__ = (IFakeTransport,)

    connLost = False
    prodReg = None
    unregProd = None
    def __init__(self):
        self.data = []
    def loseConnection(self):
        self.connLost = True
    def write(self, data):
        self.data.append(data)
    def writeSequence(self, iovec):
        self.data.extend(iovec)
    def getPeer(self):
        return "PEER"
    def registerProducer(self, producer, streaming):
        self.prodReg = (producer, streaming)
    def unregisterProducer(self):
        self.unregProd = True

class FakeProtocol:
    connMade = None
    connLost = None
    def __init__(self):
        self.data = []
    def dataReceived(self, data):
        self.data.append(data)
    def makeConnection(self, transport):
        self.connMade = transport

class DebugShimTestCase(unittest.TestCase):
    def setUp(self):
        self.s = StringIO()
        self.p = FakeProtocol()
        self.dst = tpython.DebugShimTransport(self.p, self.s)
        self.o = FakeTransport()
        self.dst.makeConnection(self.o)

    def testMakeConnection(self):
        self.assertEquals(self.s.getvalue(), "*\n")

    def testDataReceived(self):
        self.dst.dataReceived("Foo")
        self.dst.dataReceived("Bar")
        self.assertEquals(self.s.getvalue(), "*\nC 1: 'Foo'\nC 1: 'Bar'\n")
        self.assertEquals(self.p.data, ["Foo", "Bar"])

    def testLoseConnection(self):
        self.dst.loseConnection()
        self.failUnless(self.o.connLost)
        self.assertEquals(self.s.getvalue(), "*\nS 4: *\n")

    def testWrite(self):
        self.dst.write("foo")
        self.dst.write("bar")
        self.dst.write("baz")
        self.assertEquals(self.o.data, ["foo", "bar", "baz"])
        self.assertEquals(self.s.getvalue(), "*\nS 7: 'foo'\nS 7: 'bar'\nS 7: 'baz'\n")

    def testGetPeer(self):
        self.assertEquals(self.dst.getPeer(), "PEER")

    def testProducer(self):
        self.dst.registerProducer("foo", "bar")
        self.dst.unregisterProducer()
        self.assertEquals(self.o.prodReg, ("foo", "bar"))
        self.failUnless(self.o.unregProd)

    def testImplements(self):
        self.failUnless(components.implements(self.p.connMade, IFakeTransport))

class TPythonTestException(Exception):
    pass

class IteratorTest(unittest.TestCase):
    def testIterateInReactor(self):
        i = iter(range(10))
        d = tpython.iterateInReactor(i, 0)
        self.assertIdentical(unittest.deferredResult(d), i)

    def testFailureInIterateInReactor(self):
        def failer():
            yield 10
            yield 20
            raise TPythonTestException("test_tpython Test Exception")
        d = tpython.iterateInReactor(failer(), 0)
        e = unittest.deferredError(d)
        e.trap(TPythonTestException)

class HumanIntervalTestCase(unittest.TestCase):
    def testSeconds(self):
        self.assertEquals(tpython.humanInterval(0), "0 seconds")
        self.assertEquals(tpython.humanInterval(1), "1 second")
        self.assertEquals(tpython.humanInterval(30), "30 seconds")
        self.assertEquals(tpython.humanInterval(59), "59 seconds")

    def testMinutes(self):
        self.assertEquals(tpython.humanInterval(60), "1 minute")
        self.assertEquals(tpython.humanInterval(61), "1 minute, 1 second")
        self.assertEquals(tpython.humanInterval(119), "1 minute, 59 seconds")
        self.assertEquals(tpython.humanInterval(599), "9 minutes, 59 seconds")

    def testHours(self):
        self.assertEquals(tpython.humanInterval(60 * 60), "1 hour")
        self.assertEquals(tpython.humanInterval(60 * 61 + 1), "1 hour, 1 minute, 1 second")
        self.assertEquals(tpython.humanInterval(60 * 60 * 3), "3 hours")

    def testLots(self):
        self.assertEquals(tpython.humanInterval(604799), "6 days, 23 hours, 59 minutes, 59 seconds")

class LockingTestCase(unittest.TestCase):
    def testBasics(self):
        lockf = self.mktemp()
        lock = tpython.FilesystemLock(lockf)
        self.failUnless(lock.lock())
        self.failUnless(lock.clean)
        lock.unlock()
        self.failUnless(lock.lock())
        self.failUnless(lock.clean)
        lock.unlock()

    def testProtection(self):
        lockf = self.mktemp()
        lock = tpython.FilesystemLock(lockf)
        self.failUnless(lock.lock())
        self.failUnless(lock.clean)
        self.failIf(lock.lock())
        lock.unlock()

    def testBigLoop(self):
        lockf = self.mktemp()
        lock = tpython.FilesystemLock(lockf)
        self.failUnless(lock.lock())
        for i in xrange(500):
            self.failIf(lock.lock())
        lock.unlock()

    # A multiprocess test would be good here, for the sake of
    # completeness.  However, it is fairly safe to rely on the
    # filesystem to provide the semantics we require.
