# -*- test-case-name: atop.test.test_upgrade -*-

import random
import itertools

from twisted.python import components
from twisted.python import log
from twisted.python import failure
from twisted.internet import defer

from atop.store import Item, Pool, IPipe, referenced, INDEX
from atop import powerup
from atop import sched

class DeleteMe(Item):
    """I am the simplest kind of upgrade.  Use this class when you have made a
    mistake and made something persistent which you should never have stored
    anywhere, in memory or otherwise.  Use it like this:

        Your old code looks like this::

            # dumb.py
            from atop.store import Item
            class Stupid(Item):
                # do stupid stuff

        Your new code should look like this
            # dumb.py
            from atop.store import DeleteMe
            Stupid = DeleteMe

    If you are removing a subclass of Pool, make sure to use DeleteMePool
    instead.  There is currently no version of DeleteMe which works with
    Substores.
    """

    def activate(self):
        self.delete()

class DeleteMePool(Pool):
    """A version of DeleteMe that works with pools.
    """
    def activate(self):
        self.delete()

class QueueErrorPool(Pool):
    """Pool containing FailedTasks.

    Items whose process method raises an exception will end
    up in this pool, wrapped FailedTask Items.
    """

class QueuedTaskError(Exception):
    """Exception indicating one or more Items failed processing.

    @ivar queue: The Queue which had failing Items.
    """
    def __init__(self, pool):
        Exception.__init__(self, pool)
        self.queue = pool

class FailedTask(Item):
    """A wrapper around Items whose process method raises an exception.

    @ivar item: The Item which failed.
    @ivar reason: A failure created from the exception.
    """
    item = referenced

    def __init__(self, store, item, reason):
        Item.__init__(self, store)
        self.item = item
        self.reason = reason

class UpgradeQueue(Item, sched.Schedulable):
    __implements__ = (sched.Schedulable.__implements__, IPipe)

    pool = referenced
    errors = referenced
    finished = None
    failures = False

    MAXIMUM_LOCKS = 50

    strFormat = "%(poolName)r upgrader (%(poolPath)s) with %(upgradeFactory)s"

    lastPercent = 0

    def __init__(self, store, pool, upgradeFactory):
        Item.__init__(self, store)
        self.pool = pool
        self.upgradeFactory = upgradeFactory
        self.errors = QueueErrorPool(self.store)

        self.currentUID = None
        try:
            self.maxUID = pool.queryIndex('uid', backwards=True, result=INDEX).next()
        except StopIteration:
            self.maxUID = -1

    def __str__(self):
        return self.strFormat % {'poolName': self.pool.name,
                                 'poolPath': self.pool.getItemPath(),
                                 'upgradeFactory': self.upgradeFactory}

    def start(self):
        self.touch()
        assert self.finished is None, "Can only start an Upgrader once"
        self.finished = False
        self.deferred = defer.Deferred()
        self.reset(0)
        return self.deferred

    def _oneUpgrade(self):
        item = self.nextItem()
        if item is not None and item.getIndexKey(self.pool, 'uid') <= self.maxUID:
            try:
                self.upgradeFactory(self.pool, item)
            except:
                self.fail(item, failure.Failure())
        else:
            self.unschedule()
            self.finish()
            return True

    def pickDelay(self, howMany):
        return 0

    def run(self):
        self.touch()
        count = self.store.batch(self._oneUpgrade)
        if not self.finished:
            newPercent = 100.0 * self.currentUID / (self.maxUID + 1)
            if int(newPercent) != self.lastPercent:
                self.lastPercent = int(newPercent)
                log.msg("%s %d%% complete." % (self, self.lastPercent))
            self.reset(self.pickDelay(count))

    def nextItem(self):
        q = self.pool.queryIndex('uid', startKey=self.currentUID)
        try:
            item = q.next()
        except StopIteration:
            return None
        self.currentUID = item.getIndexKey(self.pool, 'uid') + 1
        return item

    def fail(self, item, failure):
        self.failures = True
        log.err(failure)
        failure.cleanFailure()
        self.errors.addItem(FailedTask(self.store, item, failure))

    def finish(self):
        if self.finished:
            # XXX What the heck?
            log.msg("Warning: double-finished %s" % (self,))
            return
        log.msg("Finished %s" % (self,))
        self.finished = True
        d = self.deferred
        del self.deferred
        if self.failures:
            d.errback(QueuedTaskError(self))
        else:
            d.callback(self)

class ItemUpgradeFactoryUpgradeQueue(UpgradeQueue):
    upgradeFactory = referenced

default = object()
class NewItemIndex:
    def __init__(self, indexName):
        self.indexName = indexName

    def __call__(self, pool, item):
        newIndexKey = item.getIndexKey(pool, self.indexName, default)
        if newIndexKey is not default:
            pool.changeKeyForItem(item, self.indexName, None, newIndexKey, notify=False)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.indexName,)

class DelItemIndex(NewItemIndex):
    def __call__(self, pool, item):
        pool.changeKeyForItem(item, self.indexName,
                              item.getIndexKey(pool,
                                               self.indexName),
                              None)

def upgradeIndex(pool, indexName):
    """Add keys to an index newly added to a Pool.

    This should be called immediately after C{pool.addIndex(indexName)}
    and before any new items are added to the pool.
    """
    return UpgradeQueue(pool.store, pool, NewItemIndex(indexName)).start()

def noop(pool, item):
    pass

def versionedUpgrade(pool):
    """Gradually load all Items from a Pool.

    This is a helper for L{twisted.persisted.styles.Versioned}.
    Object upgrade methods may have side effects which it is
    desirable to see sooner rather than later.  This function
    causes all Items in a pool to be deserialized, forcing their
    upgrade methods to run, so that those side effects may be
    seen.
    """
    return UpgradeQueue(pool.store, pool, noop).start()

class IndexCount(Item):
    __implements__ = IPipe
    pool = referenced

    currentUID = 0

    def __init__(self, pool):
        Item.__init__(self, pool.store, '(%s)' % pool.name)
        self.pool = pool
        self.pool.addPipe(self.fwd(IPipe))
        self.counts = dict.fromkeys(pool.indexes, 0)

    def __call__(self, pool, item):
        self.touch()
        self.currentUID = item.index_uid(pool) + 1
        for iname in self.counts:
            if item.getIndexKey(self.pool, iname) is not None:
                self.counts[iname] += 1

    def itemAdded(self, item):
        # We iterate in UID order.  Nothing will ever be added
        # that we won't proceed to iterate over later.
        pass

    def itemRemoved(self, item):
        if item.getIndexKey(self.pool, 'uid') < self.currentUID:
            # We counted it already.  Reverse that.
            for iname in self.counts:
                if item.getIndexKey(self.pool, iname) is not None:
                    self.counts[iname] -= 1

    def itemChanged(self, item, changeData):
        pass

def cbTransferCounts(queue):
    counter = queue.upgradeFactory
    pool = counter.pool
    def _():
        pool.removePipe(counter.fwd(IPipe))
        for iname in counter.counts:
            pool.indexCounter[iname] = counter.counts[iname]
    pool.store.transact(_)
    return counter

def ebTransferCounts(failure):
    cbTransferCounts(failure.queue)
    log.err(failure)
    return failure

def recountIndexes(pool):
    """Helper for L{store.Pool.upgradeToVersion1}.

    Counts how many Items belong to each index on the given
    pool.
    """
    ic = IndexCount(pool)
    d = ItemUpgradeFactoryUpgradeQueue(pool.store, pool, ic).start()
    d.addCallbacks(cbTransferCounts, ebTransferCounts)
    return d
