# -*- test-case-name: quotient.test.test_prefup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.persisted.styles import Versioned

from nevow.compy import Interface, Adapter

from atop.powerup import Powerup, IPowerStation
from atop.store import Pool, NameIndex, referenced
from atop.credup import Avatar
from atop.sched import CallingItem

# Persistence helper for CallingItem/navigation reset
def _callResetNav(prefRef, confRef):
    prefRef.getItem().emptyPrimaryNav()
    confRef.getItem().setupPrimaryNav()

class Avatar(Avatar, Versioned):

    persistenceVersion = 3

    def upgradeToVersion3(self):
        self.touch()
        from quotient.smarts.person import IPersonCollector
        IPersonCollector(self.store, default=None)

    def upgradeToVersion2(self):
        self.touch()
        from quotient.cleanup import IClean
        from quotient.smartenup import ISmartenup
        from quotient.cleanup import ISpambayes
        IClean(self, default=None)
        ISmartenup(self, default=None)
        ISpambayes(self, default=None)

        from quotient.prefup import IPrefs
        from quotient.configure import IConfigure
        prefs = IPrefs(self, default=None)
        config = IConfigure(self, default=None)
        if prefs is not None and config is not None:
            CallingItem(self, 0, _callResetNav,
                        prefs.referenceTo(), config.referenceTo())

        from quotient.cleanup import deduplicate
        deduplicate(self)

    def upgradeToVersion1(self):
        self.touch()
        from quotient.addressbook import IAddressBook
        # load the address book
        IAddressBook(self, default=None)

    def initialize(self):
        """Called when an atop.credup.Avatar is first turned into this Avatar.

        This would better be named `upgradeToVersion-1', but such things are
        not possible.  This method is probably of interest to no one but the
        Divmod team, since everyone else *should* have their own Avatar class,
        or be using this one directly, from day one, rather than using
        credup.credup.Avatar directly at all.  If, however, you make the same
        mistake we did and don't create your own Avatar class, this method
        might be educational to read.

        This method upgrades the Avatar as soon as the old class is replaced.
        It side-steps the normal styles.Versioned upgrade process because
        it is expected to work on an object already in memory, not one
        which has just be deserialized, and so it cannot properly depend on
        persistenceVersion being informative.
        """
        from twisted.python.components import CannotAdapt
        from twisted.web.resource import IResource as OldInterface
        from nevow.inevow import IResource as NewInterface
        try:
            comp = OldInterface(self)
            self.unsetComponent(OldInterface)
        except (KeyError, CannotAdapt):
            pass
        else:
            self.setComponent(NewInterface, comp)

class IDivmodApp(Interface):
    pass

class IMailApp(Interface):
    pass

class IVoiceApp(Interface):
    pass

class IChatApp(Interface):
    pass

class IPeopleApp(Interface):
    pass

class IAlbumApp(Interface):
    pass

class ICalendarApp(Interface):
    pass

class ITodoApp(Interface):
    pass

class IFindApp(Interface):
    pass

class IAdminApp(Interface):
    pass

class ISettingsApp(Interface):
    pass

class IApplication(Interface):
    def getApplicationNav(self):
        """
        Return a list of Items for the application's navigation.
        aka, secondary nav.
        """
        pass

    def setApplicationNav(self, listOfItems):
        """
        Take a list of items that represent this application's nav.
        """

class Application(Powerup):

    # XXX
    # Order here is crucial.  Wrong order leads to broken
    # rendering.  This should be fixed!

    __implements__ = (IApplication, Powerup.__implements__)
    powerupName = "Generic Application"
    _myInterface= IApplication

    appRoot = referenced

    def setUpPools(self, store):
        store.setComponent(self._myInterface, self)
        self.appNav = []

    def setApplicationNav(self, listOfItems):
        self.touch()
        self.appNav = [item.getRef() for item in listOfItems]

    def getApplicationNav(self):
        return [ref.getItem() for ref in self.appNav]

    def _getAppRootPool(self):
        return self.appRoot

class DivmodApp(Application):
    __implements__ = Application.__implements__, IDivmodApp
    powerupName = "Divmod"
    _myInterface = IDivmodApp
    labels = referenced
    
    persistenceVersion = 2

    def upgradeToVersion2(self):
        self.touch()
        self.labels = Pool(self.store, 'labels')
        self.labels.addIndex(NameIndex())

    def upgradeToVersion1(self):
        from atop.powerup import IPowerStation
        ps = IPowerStation(self.store, None)
        if ps is not None:
            from quotient.sumup import Summary
            ps.installPowerup(Summary)
            from quotient.configure import IConfigure
            conf = IConfigure(self.store, None)
            if conf is not None:
                conf.setupDivmodApp(conf.getPupsNPools(), self)

    def setUpPools(self, store):
        Application.setUpPools(self, store)
        self.labels = Pool(self.store, 'labels')
        self.labels.addIndex(NameIndex())


class MailAppRootPool(Pool):
    def __init__(self, *args, **kw):
        Pool.__init__(self, *args, **kw)
        self.addIndexes()

    def addIndexes(self):
        self.addIndex(NameIndex())

class MailApp(Application):
    __implements__ = Application.__implements__, IMailApp
    powerupName = "Mail"
    _myInterface = IMailApp

    persistenceVersion = 1

##    def upgradeToVersion2(self):
##        self.touch()
##        from quotient.configure import Configup
##        IPowerStation(self.store).installPowerup(Configup)

    def upgradeToVersion1(self):
        self.touch()

        from quotient.prefup import IPrefs
        from quotient.configure import IConfigure
        prefs = IPrefs(self, default=None)
        config = IConfigure(self, default=None)
        if prefs is not None and config is not None:
            prefs.emptyPrimaryNav()
            config.setupPrimaryNav()

    def setUpPools(self, store):
        Application.setUpPools(self, store)
        appRoot = MailAppRootPool(store, 'Mail Folders')
        appRoot.suggestIndex('name')
        self.store.getRootPool().addItem(appRoot)
        self.appRoot = appRoot

        # from quotient.configure import Configup
        # IPowerStation(self.store).installPowerup(Configup)

class MailAppSecNav(Adapter):
    __implements__ = Application.__implements__,
    def getApplicationNav(self):
        return IMailApp(self.original.store).getApplicationNav()


class VoiceApp(Application):
    __implements__ = Application.__implements__, IVoiceApp
    powerupName = "Voice"
    _myInterface = IVoiceApp

    def setUpPools(self, store):
        Application.setUpPools(self, store)
        appRoot = MailAppRootPool(store, 'Calls')
        self.store.getRootPool().addItem(appRoot)
        self.appRoot = appRoot

class ChatApp(Application):
    __implements__ = Application.__implements__, IChatApp
    powerupName = "Chat"
    _myInterface = IChatApp

class PeopleAppRootPool(Pool):
    pass

class PeopleApp(Application):
    __implements__ = Application.__implements__, IPeopleApp
    powerupName = "People"
    _myInterface = IPeopleApp

    persistenceVersion = 1

    def upgradeToVersion1(self):
        from quotient.addressbook import IAddressBook, AddressBook
        ab = IAddressBook(self.store, None)
        if ab is None:
            ab = IPowerStation(self.store).installPowerup(AddressBook)
        self.touch()
        self.appNav.insert(0, ab.getRef())


    def setUpPools(self, store):
        store.setComponent(self._myInterface, self)
        appRoot = PeopleAppRootPool(store, 'People')
        self.store.getRootPool().addItem(appRoot)
        self.appRoot = appRoot
        self.appNav = []

class PeopleAppSecNav(Adapter):
    __implements__ = Application.__implements__,
    def getApplicationNav(self):
        return IPeopleApp(self.original.store).getApplicationNav()

class AlbumApp(Application):
    __implements__ = Application.__implements__, IAlbumApp
    powerupName = "Album"
    _myInterface = IAlbumApp

class CalendarApp(Application):
    __implements__ = Application.__implements__, ICalendarApp
    powerupName = "Calendar"
    _myInterface = ICalendarApp

class TodoApp(Application):
    __implements__ = Application.__implements__, ITodoApp
    powerupName = "Todo"
    _myInterface = ITodoApp

class FindApp(Application):
    __implements__ = Application.__implements__, IFindApp
    powerupName = "Find"
    _myInterface = IFindApp

class AdminApp(Application):
    __implements__ = Application.__implements__, IAdminApp
    powerupName = "Admin"
    _myInterface = IAdminApp

class SettingsApp(Application):
    __implements__ = Application.__implements__, ISettingsApp
    powerupName = "Settings"
    _myInterface = ISettingsApp

class SettingsAppSecNav(Adapter):
    __implements__ = Application.__implements__, ISettingsApp
    def getApplicationNav(self):
        return ISettingsApp(self.original.store).getApplicationNav()
