
from twisted.application.service import MultiService, Application, IServiceCollection

from twisted.tap import manhole

from atop.store import Store, StorageService
from atop.sched import SchedulerService
from atop.credup import StoreRealm
from atop.service import Serviceup
from atop.powerup import IPowerStation

from quotient.signup import BoxOffice
from quotient.sysup import Sysup, ControlPanel
from quotient.meterer import MeterStoreup

from twisted.cred.credentials import CramMD5Credentials, UsernamePassword
from twisted.cred.checkers import ICredentialsChecker
from twisted.cred.portal import IRealm
from twisted.cred.portal import Portal
from twisted.cred.checkers import AllowAnonymousAccess
from twisted.spread import pb
from twisted.internet import reactor
from twisted.application.service import IService, IServiceCollection

from twisted.internet.ssl import DefaultOpenSSLContextFactory
from OpenSSL.SSL import SSLv23_METHOD, TLSv1_METHOD

from twisted.application.internet import TCPServer, SSLServer
from quotient.proto.popup import POP3Factory
from quotient.proto.imapup import IMAP4Factory
from quotient.proto.smtpin import ESMTPInFactory
from quotient.provisioner import provisionRelation
from quotient.traceup import Traceup

from quotient.web.root import AtopSite
from nevow.guard import SessionWrapper
from nevow.liveevil import LiveEvil
from quotient.proto import sip
from quotient.voiceup import AccounterFactory

from twisted.python import components

class NullRealm:
    __implements__ = IRealm

    def requestAvatar(self, avatarId, mind, *interfaces):
        raise NotImplementedError("Realm not yet initialized.")

class ReconnectingPBClientFactory(pb.PBClientFactory):

    def __init__(self, host, port, creds, client=None):
        pb.PBClientFactory.__init__(self)
        self.host = host
        self.port = port
        self.creds = creds
        self.client=client

    def clientConnectionFailed(self, connector, reason):
        pb.PBClientFactory.clientConnectionFailed(self, connector, reason)
        reactor.callLater(10, self.connect)

    def clientConnectionLost(self, connector, reason):
        pb.PBClientFactory.clientConnectionLost(self, connector, reason, reconnecting=1)
        reactor.callLater(10, self.connect)

    def connect(self):
        reactor.connectTCP(self.host, self.port, self)
        return self.login(self.creds, self.client)

class QuotientService(MultiService):
    def __init__(self, dataDirectory, fileDirectory, webLogFile, domainNames,
                 pop3Port, pop3sPort, imap4Port, imap4sPort, httpPort, httpsPort,
                 smtpPort, smtpsPort, sipPort, pbPort, certificateFile=None,
                 privateKeyFile=None, asteriskHost=None, asteriskPort=None,
                 webWrapper=None, initialTicketCount=0):
        MultiService.__init__(self)
        self.setName("Quotient")

        for arg in ('dataDirectory', 'fileDirectory', 'webLogFile',
                    'domainNames', 'pop3Port', 'pop3sPort', 'imap4Port',
                    'imap4sPort', 'httpPort', 'httpsPort', 'smtpPort',
                    'smtpsPort', 'sipPort', 'pbPort', 'certificateFile',
                    'privateKeyFile', 'asteriskHost', 'asteriskPort', 'webWrapper',
                    'initialTicketCount'):
            setattr(self, arg, locals()[arg])

    def attach(self, subservice):
        subservice.setServiceParent(self)
        return subservice

    def openStore(self):
        return Store(self.dataDirectory, self.fileDirectory)

    def privilegedStartService(self):
        # create a fake, temporary portal
        self.portal = Portal(NullRealm())
        self.webportal = Portal(NullRealm())

        # SSL housekeeping
        tlsfac = sslfac = self.sslfac = None
        if self.certificateFile is not None:
            ctxf = DefaultOpenSSLContextFactory
            cf = self.certificateFile
            pf = self.privateKeyFile
            tlsfac = ctxf(cf, pf, TLSv1_METHOD)
            sslfac = self.sslfac = ctxf(cf, pf, SSLv23_METHOD)

        rsrc = self._webWrap(SessionWrapper(self.webportal, mindFactory=LiveEvil))

        # bind all ports
        authorizers = {'CRAM-MD5': CramMD5Credentials}

        self.server("pop3", POP3Factory(self.portal, authorizers))
        self.server("imap4", IMAP4Factory(self.portal,
                                          challengers=authorizers,
                                          contextFactory=tlsfac))
        self.server("http", AtopSite(rsrc, logPath=self.webLogFile))
        self.server("smtp", ESMTPInFactory(self.webportal, authorizers, sslfac))

        return MultiService.privilegedStartService(self)

    def startService(self):
        self.store = self.openStore()

        sipd = sip.RegisterProxy(self.domainNames[0], self.sipPort)
        sipd.portal = self.webportal
        if self.asteriskHost:
            import cdrmonitor
            sipregistry = sip.ForwardingRegistry(self.domainNames, self.portal,
                                                 self.asteriskHost, self.asteriskPort)
            factory = ReconnectingPBClientFactory(self.asteriskHost, 5039,
                                                  UsernamePassword("asterisk", cdrmonitor.SECRET),
                                                  cdrmonitor.CDRClient(self.store))
            factory.connect()

            self.attach(TCPServer(5040, AccounterFactory(self.store)))
        else:
            sipregistry = sip.MultiDomainRegistry(self.domainNames)
        self.attach(sip.SIPServer(self.sipPort, sipd, sipregistry, sipregistry))
        self.attach(StorageService(self.store))
        self.store.transact(self._transactStartup)
        self.server('pb', pb.PBServerFactory(self.portal, unsafeTracebacks=1))

        return MultiService.startService(self)

    def _webWrap(self, rsrc):
        if self.webWrapper is None:
            return rsrc
        else:
            return self.webWrapper(rsrc)

    def _transactStartup(self):
        self.attach(SchedulerService(self.store))

        # install store powerups.
        storeups = [Serviceup, StoreRealm, BoxOffice,
                    ControlPanel, MeterStoreup, Traceup]
        psi = IPowerStation(self.store).installPowerup

        map(psi, storeups)

        # make sure there is a domain.
        checker = ICredentialsChecker(self.store)
        map(checker.addDomain, self.domainNames)

        realm = IRealm(self.store)
        psi(Sysup)

        # set up portal stuff
        self.webportal.realm = realm
        self.webportal.registerChecker(checker)
        self.webportal.registerChecker(AllowAnonymousAccess())

        self.portal.realm = realm
        self.portal.registerChecker(checker)


    def server(self, name, factory):
        normalPort = getattr(self, "%sPort" % name)
        self.attach(TCPServer(normalPort, factory))
        if self.sslfac is not None:
            sslPort = getattr(self, "%ssPort" % name, None)
            if sslPort:
                self.attach(SSLServer(sslPort, factory, contextFactory=self.sslfac))


from twisted.persisted.sob import IPersistable

class NotPersistable:
    __implements__ = IPersistable
    def __init__(self, original):
        self.original = original

    def setStyle(self, style):
        self.style = style

    def save(self, tag=None, filename=None, passphrase=None):
        pass


def deploy(

    # mandatory QS options

    dataDirectory, fileDirectory, webLogFile, domainNames, pop3Port, pop3sPort,
    imap4Port, imap4sPort, httpPort, httpsPort, smtpPort, smtpsPort, sipPort,
    pbPort,
    # manhole options
    manholeUser, manholePassword, manholePort,

    # Optional QS options
    certificateFile=None, privateKeyFile=None,
    asteriskHost=None, asteriskPort=None,
    webWrapper=None, initialTicketCount=0,

    # Optional deployment options
    uid=None, gid=None,
    ):

    qs = QuotientService(dataDirectory, fileDirectory, webLogFile, domainNames,
                         pop3Port, pop3sPort, imap4Port, imap4sPort, httpPort,
                         httpsPort, smtpPort, smtpsPort, sipPort, pbPort, certificateFile,
                         privateKeyFile, asteriskHost, asteriskPort, webWrapper,
                         initialTicketCount)

    a = Application("Quotient", uid=uid, gid=gid)
    a.addComponent(NotPersistable(a), ignoreClass=True)
    qs.setServiceParent(a)

    if manholeUser is not None:
        mhserver = manhole.makeService({
            'user': manholeUser,
            'password': manholePassword,
            'port': str(manholePort),
            'tracebacks': True,
            'namespace': {'service': qs}})
        mhserver.setServiceParent(a)

    return a
