# -*- test-case-name: "quotient.test.test_webui" -*-
# -i*- test-case-name: "quotient.test.test_functional" -*-

import sys
from twisted.python import log, components, util
from twisted.internet.defer import Deferred, maybeDeferred
from twisted.trial.unittest import FailTest
from quotient.functional import client

_navtags_debug = False

def debugObserver(events):
    if _navtags_debug and 'navtags' in events:
        print 'navtags: %s' % ''.join(list(events['navtags']))
log.addObserver(debugObserver)


try:
    _perIterCbs
except NameError:
    _perIterCbs = (lambda a: a, lambda b: b)

def runSeq(seq, perIterCallbacks=_perIterCbs):

    global _perIterCbs
    _perIterCbs = perIterCallbacks

    def eb(error):
        error.trap(AssertTagFailure)
        print "%s" % (error.getErrorMessage())
        raise error

    # THANK YOU JP!
    def deferIterable(dummy, f, i):
        try:
            o = i.next()
        except StopIteration:
            return
        else:
            return maybeDeferred(f, o
                                 ).addErrback(eb
                                 ).addCallbacks(*_perIterCbs
                                 ).addCallback(deferIterable, f, i)

    return deferIterable(None, lambda o: IDoMyThing(o).doYourThing(), iter(seq))


class IDoMyThing(components.Interface):
    def doYourThing(self):
        pass

class NoneHandler(components.Adapter):
    """A no-op for when None is passed in a sequence"""
    __implements__ = IDoMyThing,
    def doYourThing(self):
        log.msg(navtags='NoneHandler.doYourThing')
        return None

class FunctionHandler(components.Adapter):
    __implements__ = IDoMyThing,
    def doYourThing(self):
        log.msg(navtags='FunctionHandler.doYourThing')
        return IDoMyThing(self.original()).doYourThing()


# runSeq should go through a possibly nested list of items _depth-first_ and
# chain them using maybeDeferred 

class IterableHandler(components.Adapter):
    __implements__ = IDoMyThing,
    def doYourThing(self):
        log.msg(navtags='IterableHandler.doYourThing')
        log.msg(navtags='self.original: %s' % repr(self.original))
        return runSeq(self.original, _perIterCbs)


class LinkTag:
    def __init__(self, id=''):
        self.id = id
    def __repr__(self):
        return '%s object <%s>: %s\n' % (self.__class__, hex(id(self)), repr(str(self.id)))

class NavigableLink(components.Adapter):
    __implements__ = IDoMyThing,
    def doYourThing(self):
        log.msg(navtags='NavigableLink.doYourThing')
        return client.page.getLink(self.original.id).follow()


class URLTag:
    def __init__(self, href="", *a, **kw):
        self.href, self.a, self.kw = href, a, kw
    def __repr__(self):
        return '<%s object %s  href: %s  a %s  kw: %s>' % (self.__class__, hex(id(self)), self.href, str(self.a), str(self.kw))

class NavigableURL(components.Adapter):
    __implements__ = IDoMyThing,
    def doYourThing(self):
        log.msg(navtags='NavigableURL.doYourThing')
        o = self.original
        return client.get(o.href + '!'.join(o.a), o.kw)


class FormTag:
    def __init__(self, id='', **kw):
        self.id, self.kw = id, kw
    def __repr__(self):
        return '<%s object at %s  id: %s  kw: %s>' % (self.__class__, hex(id(self)), str(self.id), str(self.kw))

class NavigableForm(components.Adapter):
    __implements__ = IDoMyThing,
    def doYourThing(self):
        log.msg(navtags='NavigableForm.doYourThing')
        o = self.original
        return client.page.getForm(o.id).submit(o.kw)


class CallTag:
    """use this to get a method called with the given args and kwargs
    you can use this instead of lambdas"""
    def __init__(self, func, *a, **kw):
        self.func, self.a, self.kw = func, a, kw

class CallTagAdapter(components.Adapter):
    __implements__ = IDoMyThing,
    def doYourThing(self):
        log.msg(navtags='FunctionTagAdapter.doYourThing')
        o = self.original
        return IDoMyThing(o.func(*o.a, **o.kw)).doYourThing()


class IAssertTag(components.Interface):
    pass

class AssertTag(object):
    def __init__(self, *a):
        self.a = a
    def __repr__(self):
        return '<%s object at %s  a: %s>' % (self.__class__, hex(id(self)), str(self.a))


def _unpackArgs(numArgs, L):
    if len(L) == numArgs - 1:
        return tuple(L + (None,))
    elif len(L) == numArgs:
        return tuple(L)
    raise RuntimeError("array wrong size, numArgs was %s but len(L) was %s" % (numArgs, len(L)))

class AssertTagFailure(FailTest):
    pass

class FailIf(AssertTag):
    __implements__ = IDoMyThing, 
    name = 'FailIf'
    def doYourThing(self):
        cond, msg = _unpackArgs(2, self.a)
        if cond:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, msg)

class FailUnless(AssertTag):
    __implements__ = IDoMyThing,
    name = 'FailUnless'
    def doYourThing(self):
        cond, msg = _unpackArgs(2, self.a)
        if not cond:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, msg)

class FailUnlessEqual(AssertTag):
    __implements__ = IDoMyThing,
    name = 'FailUnlessEqual'
    def doYourThing(self):
        first, second, msg = _unpackArgs(3, self.a)
        if not first == second:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, (msg or '%r != %r' % (first, second)))
            
class FailUnlessIdentical(AssertTag):
    __implements__ = IDoMyThing,
    name = 'FailUnlessIdentical'
    def doYourThing(self):
        first, second, msg = _unpackArgs(3, self.a)
        if first is not second:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, (msg or '%r is not %r' % (first, second)))

class FailIfIdentical(AssertTag):
    __implements__ = IDoMyThing,
    name = 'FailIfIdentical'
    def doYourThing(self):
        first, second, msg = _unpackArgs(3, self.a)
        if first is second:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, (msg or '%r is %r' % (first, second)))

class FailIfEqual(AssertTag):
    __implements__ = IDoMyThing,
    name = 'FailIfEqual'
    def doYourThing(self):
        first, second, msg = _unpackArgs(3, self.a)
        if not first != second:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, (msg or '%r == %r' % (first, second)))

class FailUnlessIn(AssertTag):
    __implements__ = IDoMyThing,
    name = 'FailUnlessIn'
    def doYourThing(self):
        containee, container, msg = _unpackArgs(3, self.a)
        if containee not in container:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, (msg or "%r not in %r" % (containee, container)))

class FailIfIn(AssertTag):
    __implements__ = IDoMyThing,
    name = 'FailIfIn'
    def doYourThing(self):
        containee, container, msg = _unpackArgs(3, self.a)
        if containee in container:
            raise AssertTagFailure, "%s failed with msg: %s" % (self.name, (msg or "%r in %r" % (containee, container)))

class StringAssertionBase(AssertTag):
    name = None
    def _errorMsg(self):
        from os.path import abspath
        absUrlTrailPath = abspath(client.urlTrail.path)
        return '\n'.join(["\n\n%s failed with msg: %s" % (self.name, "'%s' not found in page" % (self.str)),
                          "That page was saved to %s" % client.urlTrail.filename,
                          "the urltrail file is saved in %s" % client.urlTrail.path,
                          "which you can view in your web browser at url: file://%s\n\n" % absUrlTrailPath ])

class AssertString(StringAssertionBase):
    __implements__ = IDoMyThing,
    name = 'AssertString'
    def doYourThing(self):
        self.str, msg = _unpackArgs(2, self.a)
        if self.str not in client.page.domstr:
            raise AssertTagFailure, (msg or self.errorMsg)
    errorMsg = property(StringAssertionBase._errorMsg) # because python sucks!


class AssertNotString(StringAssertionBase):
    __implements__ = IDoMyThing,
    name = 'AssertNotString'
    def doYourThing(self):
        self.str, msg = _unpackArgs(2, self.a)
        if self.str in client.page.domstr:
            raise AssertTagFailure, (msg or self.errorMsg)
    errorMsg = property(StringAssertionBase._errorMsg) # because python sucks!


assertString = failUnlessString = AssertString
assertNotString = failIfString = AssertNotString
assertEqual = assertEquals = failUnlessEqual = FailUnlessEqual
assertNotEqual = assertNotEquals = failIfEqual =  FailIfEqual
assert_ = failUnless = FailUnless
failIfEquals = failIfEqual = FailIfEqual
assertIdentical = failUnlessIdentical = FailUnlessIdentical
assertNotIdentical = failIfIdentical = FailIfIdentical
assertIn = failUnlessIn = FailUnlessIn
assertNotIn = failIfIn = FailIfIn
failIf = FailIf

assertions = ['assertEquals', 'assertEqual', 'failUnlessEqual', 'assertNotEquals', 'assertNotEqual',
              'failIfEqual', 'assert_', 'failUnless', 'assertIdentical', 'failUnlessIdentical',
              'assertNotIdentical', 'assertIn', 'failUnlessIn', 'assertNotIn', 'failIfIdentical',
              'failIfIn', 'assertString', 'failUnlessString', 'assertNotString', 'failIfString', 'failIf']

call, link, assrt, url, form = CallTag, LinkTag, AssertTag, URLTag, FormTag 


__all__ = ['runSeq', 'call', 'link', 'assrt', 'url', 'form', 'IDoMyThing'] + assertions
