# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


from twisted.python import components
from twisted.python import log
from twisted.application.service import IService, IServiceCollection

from atop import powerup
from atop.service import Serviceup

from quotient import stats

class IMeter(components.Interface):
    def amountUsed(self, name):
        """Retrieve the amount of the named resource used.
        """

    def use(self, resource, amount):
        """Use up an amount of a non-releasable resource.
        """

    def acquire(self, resource, amount):
        """Record the acquisition of an amount of the named resource.

        It is expected that this resource chunk may be released at some
        future point in time.
        """

    def release(self, resource, amount):
        """Record the release of an amount of the named resource.

        This must have been acquired previously.
        """

DISK_USAGE = "disk-usage"
BANDWIDTH_UP = "upstream-bandwidth"
BANDWIDTH_DOWN = "downstream-bandwidth"
CPU_USAGE = "cpu-usage"

class IMeterStoreup(components.Interface):
    """Funkify your Store"""

class MeterStoreup(powerup.Storeup):
    __implements__ = (IMeterStoreup, IService)

    storeDependencies = [Serviceup]
    powerupName = "Resource Meter"

    running = False

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        s = powerup.IPowerStation(self.store).installPowerup(Serviceup)
        self.name = self.powerupName
        s.addService(self)

    def setUpPools(self, store):
        store.setComponent(IMeterStoreup, self)
        IServiceCollection(self.store).addService(self)

    def privilegedStartService(self):
        raise AssertionError("privilegedStartService *cannot* be called!")

    def startService(self):
        self.touch()
        self.running = True
        log.addObserver(self.observe)

    def stopService(self):
        self.touch()
        self.running = False
        log.removeObserver(self.observe)

    def observe(self, events):
        if events['message']:
            return
        for pfx in ('permissioner', 'provisioner', 'storage'):
            f = events.get(pfx)
            if f:
                getattr(self, '_'.join((pfx, f)), lambda _: None)(events)
                return

    def permissioner_check(self, events):
        pass

    def provisioner_deplete(self, events):
        m = IMeter(events['avatar'], default=None)
        if m:
            m.use(events['name'], events['amount'])

    def provisioner_refuel(self, events):
        pass

    def storage_used(self, events):
        m = IMeter(events['avatar'], default=None)
        if m:
            m.use(DISK_USAGE, events['amount'])

class Meterup(powerup.Powerup):
    __implements__ = (powerup.Powerup.__implements__, IMeter)

    powerupName = "Resource Meter"

    def setUpPools(self, avatar):
        self.resources = stats.ResourceMeter()
        avatar.setComponent(IMeter, self)

    def amountUsed(self, resource):
        return self.resources.used(resource)

    def acquire(self, resource, amount):
        self.touch()
        self.resources.use(resource, amount)
    use = acquire

    def release(self, resource, amount):
        self.touch()
        self.resources.release(resource, amount)


from nevow import tags
import formless

from quotient.web.root import Base
from quotient import taglib

class IMeterDisplay(formless.TypedInterface):
    pass

class MeterDisplay(Base):
    __implements__ = (Base.__implements__, IMeterDisplay)

    def render_resources(self, context, data):
        f = self.original.amountUsed
        rnames = self.original.resources.resourceNames()
        return [tags.p[n, f(n)] for n in rnames]

    document = taglib.page(title="Resource Meter")[
        render_resources,
        ]
