# -*- test-case-name: quotient.test.test_imap -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
import bisect
import operator
import weakref
import time
import rfc822

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from atop import store, util, tpython

from twisted.python import reflect
from twisted.python import log
from twisted.python.util import OrderedDict
from twisted.protocols import imap4
from twisted.python import components
from twisted.internet import protocol
from twisted.internet import defer

from quotient import mimemessage
from quotient import provisioner
from quotient import cleanup
from quotient.proto import mboxlist
from quotient.application import MailApp, IMailApp

MESSAGE_FLAGS = [
    '\\Answered',
    '\\Deleted',
    '\\Draft',
    '\\Flagged',
    '\\Recent',
    '\\Seen',
    ]

def poolByName(parent, name):
    try:
        return parent.getIndexItem('imapname', name, default=None)
    except store.NoSuchIndex:
        for smeth in (str.lower, str.upper, str.capitalize, str.title):
            try:
                return parent.getIndexItem('name', smeth(name))
            except KeyError:
                pass
        return None

class IMAPAccountUp(provisioner.ProvisionedPowerup):
    __implements__ = (imap4.IAccount, provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable)
    powerupName = 'IMAP4 Server'

    dependencies = [cleanup.Cleanup, MailApp]

    H_DELIM = '/'

    startCost = 1
    tankID = provisioner.ETank.INCOMING_MAIL

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        del self.rootRef

    # Cred haxor
    def logout(self):
        pass

    def setUpPools(self, substore):
        provisioner.ProvisionedPowerup.setUpPools(self, substore)

        self.subscriptions = {}

        substore.setComponent(imap4.IAccount, self)
        self.greyedOut = True
        self.invisible = True

    def _createStep(self, part, old, factory):
        try:
            pool = poolByName(old, part)
        except store.NoSuchIndex:
            raise imap4.MailboxException("Cannot create folders there")
        else:
            if pool is None:
                pool = factory(self.store, part)
                old.addItem(pool)
            return pool

    def _createPool(self, old, path, pathspec):
        newPool = old
        path = list(path)
        lastPart = path.pop()
        for part in path:
            old = self._createStep(part, old, store.Pool)
        try:
            if poolByName(old, lastPart) is not None:
                raise imap4.MailboxCollision(pathspec)
            else:
                self._createStep(lastPart, old, mimemessage.EmailPool)
        except store.NoSuchIndex:
            raise imap4.MailboxException("Cannot create folders there")

    def create(self, pathspec):
        # XXX - Get a case insensitive version of NameIndex
        path = pathspec.lower()
        path = path.split(self.H_DELIM)
        oldPool = IMailApp(self.store).appRoot
        self.store.transact(self._createPool, oldPool, path, pathspec)
        return True

    def addMailbox(self, name, mbox = None):
        raise RuntimeError("Don't call this")


    def select(self, name, rw=True):
        parts = name.split(self.H_DELIM)
        if parts[0] == 'inbox':
            del parts[0]
            root = cleanup.IClean(self.store).cleanPool
        else:
            root = IMailApp(self.store).appRoot
        def poolByHierarchicalName(parts, pool):
            for p in parts:
                pool = poolByName(pool, p)
                if pool is None:
                    return None
            return self.store.transact(deferredOpenPool, self.store, pool, rw)
        try:
            return self.store.transact(poolByHierarchicalName, parts, root)
        except store.NoSuchIndex:
            # raise imap4.MailboxException("Cannot select that folder")
            return None

    def delete(self, name):
        raise imap4.MailboxException("Use the web UI to delete things, please.")

    def rename(self, oldname, newname):
        raise imap4.MailboxException("Use the web UI to rename things, please.")

    def isSubscribed(self, name):
        return name.lower() in self.subscriptions

    def subscribe(self, name):
        def _():
            self.touch()
            self.subscriptions[name.lower()] = True
        self.store.transact(_)
        return True

    def unsubscribe(self, name):
        def _():
            self.touch()
            del self.subscriptions[name.lower()]
        try:
            self.store.transact(_)
        except KeyError:
            raise imap4.MailboxException("Not subscribed to that mailbox.")
        return True

    def listMailboxes(self, ref, wildcard):
        def allThatJazz():
            result = []
            filterExpr = imap4.wildcardToRegexp(wildcard, None)
            q = mboxlist.listMailboxes(self.store.store, IMailApp(self.store).appRoot)
            for (name, pool) in q:
                if filterExpr.match(name):
                    result.append((name, pool))
                yield None
            yield result
        def postProcess(pools):
            done = []
            if not pools:
                return done
            for (name, pool) in pools:
                mbox = self.store.transact(deferredOpenPool, self.store, pool, False, minimal=True)
                if mbox is None:
                    continue
                done.append((name, mbox))
            return done
        return imap4.iterateInReactor(allThatJazz()
            ).addCallback(postProcess
            )

    def getPersonalNamespaces(self):
        return [["", None]]

    def getSharedNamespaces(self):
        return None

    def getUserNamespaces(self):
        return None

def deferredOpenPool(avatar, pool, rw, minimal=False):
    if isinstance(pool, mimemessage.EmailPool):
        if minimal:
            return _mailboxInfo
        q = pool.queryIndex(result=store.INDEX_AND_STORE_ID)
        return util.asyncPoolQuery(q
            ).addCallback(lambda items: Poolbox(avatar, items, pool, rw)
            )
    elif 'imapname' not in pool.indexes:
        return None
    else:
        if minimal:
            return _folderInfo
        return util.asyncPoolQuery(pool.queryIndex('name')
            ).addCallback(lambda items: FolderMailbox(avatar, items, pool, rw)
            )

class FolderInfo:
    try:
        __implements__ = imap4.IMailboxInfo
    except AttributeError:
        pass

    # All our folders have the same info atm
    def getFlags(self):
        return ['\\NoSelect']

    def getHierarchicalDelimiter(self):
        return "/"
_folderInfo = FolderInfo()

class MailboxInfo:
    try:
        __implements__ = imap4.IMailboxInfo
    except AttributeError:
        pass

    def getFlags(self):
        return Poolbox.FLAGS.keys()

    def getHierarchicalDelimiter(self):
        return "/"
_mailboxInfo = MailboxInfo()

class FolderMailbox(components.Adapter):
    __implements__ = (imap4.IMailbox, store.IPipe)

    def __init__(self, avatar, folders, original, rw):
        components.Adapter.__init__(self, original)
        self.avatar = avatar
        self.rw = rw

    # IPipe
    def itemAdded(self, item):
        pass

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, changeData):
        pass

    # IMailbox (part of it anyway)
    def getHierarchicalDelimiter(self):
        return "/"

    def getUIDValidity(self):
        return self.original.storeID

    def isWriteable(self):
        return self.rw

    def getFlags(self):
        return ['\\NoSelect']

    def getMessageCount(self):
        return 0

    def getUnseenCount(self):
        return 0

    def getRecentCount(self):
        return 0

    def requestStatus(self, names):
        return imap4.statusRequestHelper(self, names)

SEEN = mimemessage.MIMEMessage.SEEN

_faces = imap4.IMailboxInfo, imap4.IMailbox, imap4.ICloseableMailbox, imap4.IMessageCopier

class Poolbox:

    __implements__ = _faces + (store.IPipe,)

    FLAGS = {
        '\\Seen': mimemessage.MIMEMessage.SEEN,
        '\\Answered': mimemessage.MIMEMessage.ANSWERED,
        '\\Forwarded': mimemessage.MIMEMessage.FORWARDED,
        '\\Redirected': mimemessage.MIMEMessage.REDIRECTED,
        '\\Deleted': mimemessage.MIMEMessage.DELETED,
    }

    exists = None
    recent = None

    ITEM_TYPES = (mimemessage.MIMEMessage)

    def __init__(self, avatar, openInfo, original, rw):
        """
        @param openInfo: a list of two-tuples of (uid, storeID)
        """
        self.avatar = avatar
        self.original = original

        self.uids = []
        self.refs = []
        self.items = {}

        for (uid, storeID) in openInfo:
            self.addItemRef(storeID, uid)

        self.listeners = []
        self.rw = rw
        self.original.addPipe(self, ephemeral=True)
        self.unseen = self.getMessageCount() - original.getFlagCount(SEEN)

    def skip(self, item):
        return not isinstance(item, self.ITEM_TYPES)

    # ICloseableMailbox

    def close(self):
        self.original.removePipe(self, ephemeral=True)
        self.uids = self.items = self.refs = ()

    # IPipe implementation

    def itemAdded(self, item):
        if not self.skip(item):
            self.addItem(item)

            if not item.getIndexKey(self.original, 'flags', 0) & SEEN:
                self.unseen += 1
            exists = self.getMessageCount()
            recent = self.getRecentCount()
            for L in self.listeners:
                L.newMessages(exists, recent)

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, changeData):
        if isinstance(changeData, dict) and 'flags' in changeData:
            if changeData['oldFlags'] & SEEN:
                if not changeData['flags'] & SEEN:
                    self.unseen += 1
            else:
                if changeData['flags'] & SEEN:
                    self.unseen -= 1
            try:
                seqNum = self.seqByItem(item)
            except ValueError:
                pass
            else:
                d = {seqNum: explodeFlags(self.FLAGS,
                                          item.getIndexKey(self.original, 'flags', 0))}
                for L in self.listeners:
                    L.flagsChanged(d)

    # Helper functions
    def addItem(self, item):
        self.addItemRef(self.refByItem(item), item.index_uid(self.original))
        self.cacheItem(item)

    def addItemRef(self, itemRef, uid):
        self.refs.append(itemRef)
        self.uids.append(uid)

    def removeItem(self, item):
        seq = self.seqByItem(item)
        # try:
        #     del self.items[self.refByItem(item)]
        # except KeyError:
        #     pass
        del self.uids[seq - 1]
        del self.refs[seq - 1]

    def cacheItem(self, item):
        # This is too expensive currently.
        # self.items[self.refByItem(item)] = item
        return item

    def itemByRef(self, ref):
        return self.avatar.getItemByID(ref)

    def refByItem(self, item):
        return item.storeID

    def itemByUID(self, uid):
        return self.cacheItem(self.itemByRef(self.refs[self.uids.index(uid - 1)]))

    def itemBySeq(self, seq):
        return self.cacheItem(self.itemByRef(self.refs[seq - 1]))

    def uidByItem(self, item):
        return self.uids[self.refs.index(self.refByItem(item))] + 1

    def uidBySeq(self, seq):
        return self.uids[seq - 1] + 1

    def seqByItem(self, item):
        return self.refs.index(self.refByItem(item)) + 1

    def seqByUID(self, uid):
        return self.uids.index(uid - 1) + 1

    def itermessages(self):
        for ref in self.refs:
            yield self.itemByRef(ref)

    # IMailbox implementation

    def getHierarchicalDelimiter(self):
        return "/"

    def getUIDValidity(self):
        return self.original.storeID

    def getUIDNext(self):
        return self.original.nextUID(allocate=False)

    def getUID(self, message):
        """Get the UID for a given sequence number.

        Sequence numbers begin at one.
        """
        return self.uidBySeq(message)

    def getFlags(self):
        return self.FLAGS.keys()

    def getMessageCount(self):
        return len(self.refs)

    def getRecentCount(self):
        return 0

    def getUnseenCount(self):
        return self.unseen

    def isWriteable(self):
        return self.rw

    def destroy(self):
        raise imap4.MailboxException("Please delete folders using the web UI.")

    def requestStatus(self, names):
        return imap4.statusRequestHelper(self, names)

    def addListener(self, listener):
        self.listeners.append(listener)

    def removeListener(self, listener):
        self.listeners.remove(listener)

    def addMessage(self, message, flags = (), date = None):
        def cbFeedFile(msg):
            f = 0
            if '\\Seen' in flags:
                f |= msg.SEEN
            if '\\Answered' in flags:
                f |= msg.ANSWERED
            def doIt(pool, msg, flags):
                msg.changeFlags(self.original, flags)
                pool.addItem(msg)
            self.avatar.transact(doIt, self.original, msg, f)
        mr = mimemessage.MIMEMessageReceiver(self.avatar, cbFeedFile, True)
        return mr.feedFile(message).addCallback(self.uidByItem)

    def expunge(self):
        delete = []
        def collectDeleted(rate=5):
            limit = rate
            for (i, msg) in enumerate(self.itermessages()):
                if msg.getIndexKey(self.original, 'flags', 0) & msg.DELETED:
                    delete.append((i, msg))
                limit -= 1
                if limit == 0:
                    limit = rate
                    yield None
        def removeDeleted():
            delete.reverse()
            for (i, msg) in delete:
                try:
                    self.original.removeItem(msg)
                except KeyError:
                    # Maybe someone beat us to it, no big deal
                    pass
                self.removeItem(msg)
            ids = [i + 1 for (i, msg) in delete]
            ids.sort()
            ids.reverse()
            return ids
        g = collectDeleted()
        d = tpython.iterateInReactor(g)
        d.addCallback(lambda r: self.avatar.transact(removeDeleted))
        return d

    def _getnext(self, last):
        # for i in range(1, self.getMessageCount() + 1):
        #     if self.uidBySeq(i) > last:
        #         return self.uidBySeq(i)
        # raise StopIteration()

        # XXX - Abstraction violation - but it's fast!
        idx = bisect.bisect(self.uids, last - 1)
        if idx >= len(self.uids):
            raise StopIteration()
        return self.uids[idx] + 1

    def _seqAndUIDByEither(self, msgset, uid):
        if not self.getMessageCount():
            return

        if uid:
            msgset.last = self.uidBySeq(self.getMessageCount()) + 1
            msgset.getnext = self._getnext
            def f(uid):
                """Return a sequence number and a UID, given a UID.

                Sequence numbers start at 1.
                """
                return self.seqByUID(uid), uid
        else:
            msgset.last = self.getMessageCount()
            def f(seq):
                """Return a sequence number and a UID, given a sequence number.

                Sequence numbers start at 1.
                """
                return seq, self.uidBySeq(seq)

        for msg in msgset:
            try:
                yield f(msg)
            except (IndexError, ValueError):
                log.err()
                log.msg("Erroneous index requested: %s" % (msg,))

    def fetch(self, msgset, uid):
        for s, u in self._seqAndUIDByEither(msgset, uid):
            yield s, Message(self.original, self.itemByUID(u), u)

    def store(self, msgset, flags, mode, uid):
        if not self.isWriteable():
            raise imap4.ReadOnlyMailbox("Cannot store")

        state = []

        flags = reduce(operator.or_, [self.FLAGS.get(f, 0) for f in flags])

        for (seqNum, uid) in self._seqAndUIDByEither(msgset, uid):
            msg = self.itemByUID(uid)
            msgFlags = msg.getIndexKey(self.original, 'flags', 0)
            seen = msg.SEEN & msgFlags
            answered = msg.ANSWERED & msgFlags

            f = 0
            if mode == -1:
                for bv in self.FLAGS.values():
                    if bv & flags & msgFlags:
                        f |= bv
                if f:
                    state.append((msg, f, False, seqNum))
            elif mode == 0:
                state.append((msg, msgFlags, False, seqNum))
                state.append((msg, flags, True, seqNum))
            elif mode == 1:
                for bv in self.FLAGS.values():
                    if bv & flags & ~msgFlags:
                        f |= bv
                if f:
                    state.append((msg, f, True, seqNum))

        if state:
            def _():
                # Policy decision: The deleted flag will be set and unset
                # only for this folder/pool.  All other flags will be set
                # for all pools to which this message belongs.
                for (msg, flags, set, seqNum) in state:
                    if flags & msg.DELETED:
                        flags = flags & ~msg.DELETED
                        if flags:
                            msg.changeFlags(None, flags, set)
                        msg.changeFlags(self.original, msg.DELETED, set)
                    else:
                        msg.changeFlags(None, flags, set)
            self.avatar.transact(_)
        # We are to return a dict mapping message ids/uids to lists
        # of flags so that IMAP4Server can notify the client.  However,
        # our definition of changeFlag notifies all the relevant listeners,
        # so allowing IMAP4Server to perform the notification results in
        # everything being sent twice.  Return an empty dictionary to silence
        # it, instead.
        return {}

    # IMessageCopier implementation
    def copy(self, msg):
        assert isinstance(msg, Message)
        def _():
            self.original.addItem(msg.msg)
            return msg.msg.index_uid(self.original)
        return self.avatar.transact(_)

class _MsgMixin:
    def getHeaders(self, negate, *names):
        r = OrderedDict()
        if negate:
            opt = dict.fromkeys(map(str.lower, names))
            for (n, v) in self.msg.items():
                n = n.lower()
                if n not in opt:
                    r[n] = v
        else:
            for n in names:
                n = n.lower()
                v = self.msg[n]
                if v:
                    r[n] = v
        return r

    def isMultipart(self):
        return self.msg.is_multipart()

    def getSubPart(self, part):
        return SubMessage(self.msg.children[part])

    def getSize(self):
        return self.msg.size

def explodeFlags(flagDict, flagBv):
    return [k for (k, v) in flagDict.items() if flagBv & v]

class Message(_MsgMixin):

    __implements__ = (imap4.IMessage, imap4.IMessageFile)

    def __init__(self, pool, msg, uid):
        assert isinstance(msg, mimemessage.MIMEMessage)
        self.pool = pool
        self.msg = msg
        self.uid = uid

    def __str__(self):
        return '<IMAP4 Message %s>' % (tuple(self.msg.getItemPath()),)

    def __repr__(self):
        return str(self)

    def getUID(self):
        return self.uid

    def getFlags(self):
        return explodeFlags(Poolbox.FLAGS, self.msg.getIndexKey(self.pool, 'flags', 0))

    def getInternalDate(self):
        date = self.msg['x-divmod-processed']
        if date is None:
            log.msg("Missing x-divmod-processed header on %s" % (self.msg.getItemPath(),))
            date = self.msg['date'] or rfc822.formatdate()
        return date

    def getBodyFile(self):
        f = self.msg.openFile()
        f.seek(self.msg.bodyOffset)
        return f

    def open(self):
        # Neither Mozilla nor Thunderbird in Debian can handle
        # \n line-endings.
        return StringIO(self.msg.openFile().read().replace('\n', '\r\n'))

class SubMessage(_MsgMixin):

    __implements__ = (imap4.IMessagePart,)

    def __init__(self, mimepart):
        self.msg = mimepart

    def getBodyFile(self):
        return StringIO(self.msg.get_payload(decode=False))

    def getSize(self):
        return len(self.msg.openFile().read())

class IMAP4Server(imap4.IMAP4Server):
    def connectionMade(self):
        imap4.IMAP4Server.connectionMade(self)
        if self.transport.getHost()[0] == 'SSL':
            self.challengers = self.challengers.copy()
            self.challengers['LOGIN'] = imap4.LOGINCredentials
            self.challengers['PLAIN'] = imap4.PLAINCredentials

    def connectionLost(self, reason):
        imap4.IMAP4Server.connectionLost(self, reason)

    def newMessages(self, exists, recent):
        if hasattr(imap4.IMAP4Server, 'do_IDLE'):
            imap4.IMAP4Server.newMessages(self, exists, recent)

    def _listWork(self, tag, ref, mbox, sub, cmdName):
        mbox = self._parseMbox(mbox)
        defer.maybeDeferred(self.account.listMailboxes, ref, mbox
            ).addCallback(self._cbListWork, tag, sub, cmdName
            ).addErrback(self._ebListWork, tag
            )

    def _cbListWork(self, mailboxes, tag, sub, cmdName):
        for (name, box) in mailboxes:
            if not sub or self.account.isSubscribed(name):
                flags = box.getFlags()
                delim = box.getHierarchicalDelimiter()
                resp = (imap4.DontQuoteMe(cmdName), map(imap4.DontQuoteMe, flags), delim, name)
                self.sendUntaggedResponse(imap4.collapseNestedLists(resp))
        self.sendPositiveResponse(tag, '%s completed' % (cmdName,))

    def _ebListWork(self, failure, tag):
        self.sendBadResponse(tag, "Server error encountered while listing mailboxes.")
        log.err(failure)

    auth_LIST = (_listWork, imap4.IMAP4Server.arg_astring, imap4.IMAP4Server.arg_astring, 0, 'LIST')
    select_LIST = auth_LIST

    auth_LSUB = (_listWork, imap4.IMAP4Server.arg_astring, imap4.IMAP4Server.arg_astring, 1, 'LSUB')
    select_LSUB = auth_LSUB

    def _cbSelectWork(self, mbox, cmdName, tag):
        if mbox is not None and '\\noselect' in [s.lower() for s in mbox.getFlags()]:
            self.sendNegativeResponse(tag, 'Mailbox cannot be selected')
            return
        imap4.IMAP4Server._cbSelectWork(self, mbox, cmdName, tag)

class IMAP4Factory(protocol.ServerFactory):
    protocol = IMAP4Server
    debug = False
    logfile = None

    def __init__(self, portal, challengers = None, contextFactory = None):
        self.portal = portal
        if challengers is None:
            challengers = {}
        self.challengers = challengers
        self.contextFactory = contextFactory

    def startFactory(self):
        self.connectedProtocols = weakref.WeakKeyDictionary()

    def stopFactory(self):
        for p in self.connectedProtocols:
            try:
                p.transport.loseConnection()
            except:
                log.err()
        self.connectedProtocols = None

    def buildProtocol(self, addr):
        p = self.protocol(self.challengers, self.contextFactory)
        p.portal = self.portal
        p.factory = self
        self.connectedProtocols[p] = True
        if self.debug:
            if not self.logfile:
                self.logfile = file('imap4.debug', 'w')
            p = tpython.DebugShimTransport(p, self.logfile, 1024 * 1024)
        return p

# Monkey patch some fixes
def unquote(s):
    if s[0] == s[-1] == '"':
        return s[1:-1]
    return s

def getBodyStructure(msg, extended=False):
    # XXX - This does not properly handle multipart messages
    # BODYSTRUCTURE is obscenely complex and criminally under-documented.

    attrs = {}
    headers = 'content-type', 'content-id', 'content-description', 'content-transfer-encoding'
    headers = msg.getHeaders(False, *headers)
    mm = headers.get('content-type')
    if mm:
        mm = ''.join(mm.splitlines())
        mimetype = mm.split(';')
        if mimetype:
            type = mimetype[0].split('/', 1)
            if len(type) == 1:
                major = type[0]
                minor = None
            elif len(type) == 2:
                major, minor = type
            else:
                major = minor = None
            attrs = dict([x.strip().lower().split('=', 1) for x in mimetype[1:]])
        else:
            major = minor = None
    else:
        major = minor = None


    size = str(msg.getSize())
    unquotedAttrs = [(k, unquote(v)) for (k, v) in attrs.iteritems()]
    result = [
        major, minor,                       # Main and Sub MIME types
        unquotedAttrs,                      # content-type parameter list
        headers.get('content-id'),
        headers.get('content-description'),
        headers.get('content-transfer-encoding'),
        size,                               # Number of octets total
    ]

    if major is not None:
        if major.lower() == 'text':
            result.append(str(imap4.getLineCount(msg)))
        elif (major.lower(), minor.lower()) == ('message', 'rfc822'):
            contained = msg.getSubPart(0)
            result.append(imap4.getEnvelope(contained))
            result.append(getBodyStructure(contained, False))
            result.append(str(imap4.getLineCount(contained)))

    if not extended or major is None:
        return result

    if major.lower() != 'multipart':
        headers = 'content-md5', 'content-disposition', 'content-language'
        headers = msg.getHeaders(False, *headers)
        disp = headers.get('content-disposition')

        # XXX - I dunno if this is really right
        if disp:
            disp = disp.split('; ')
            if len(disp) == 1:
                disp = (disp[0].lower(), None)
            elif len(disp) > 1:
                disp = (disp[0].lower(), [x.split('=') for x in disp[1:]])

        result.append(headers.get('content-md5'))
        result.append(disp)
        result.append(headers.get('content-language'))
    else:
        result = [result]
        try:
            i = 0
            while True:
                submsg = msg.getSubPart(i)
                result.append(getBodyStructure(submsg))
                i += 1
        except IndexError:
            result.append(minor)
            result.append(attrs.items())

            # XXX - I dunno if this is really right
            headers = msg.getHeaders(False, 'content-disposition', 'content-language')
            disp = headers.get('content-disposition')
            if disp:
                disp = disp.split('; ')
                if len(disp) == 1:
                    disp = (disp[0].lower(), None)
                elif len(disp) > 1:
                    disp = (disp[0].lower(), [x.split('=') for x in disp[1:]])

            result.append(disp)
            result.append(headers.get('content-language'))

    return result
imap4.getBodyStructure = getBodyStructure
