# -*- test-case-name: quotient.test.test_pop3client -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import re
import md5

from twisted.python import log
from twisted.internet import defer
from twisted.protocols import basic

defer.debug = True

OK = '+OK'
ERR = '-ERR'

class POP3ClientError(Exception):
    pass

class InsecureAuthenticationDisallowed(POP3ClientError):
    pass

class ServerErrorResponse(POP3ClientError):
    pass

class LineTooLong(POP3ClientError):
    pass

class _ListSetter:
    def __init__(self, L):
        self.L = L
    def setitem(self, (item, value)):
        diff = item - len(self.L) + 1
        if diff > 0:
            self.L.extend([None] * diff)
        self.L[item] = value

def _listXform(line):
    index, size = line.split(None, 1)
    return int(index) - 1, int(size)

def _uidXform(line):
    index, uid = line.split(None, 1)
    return int(index) - 1, uid

def _codeStatusSplit(line):
    parts = line.split(' ', 1)
    if len(parts) == 1:
        return parts[0], ''
    return parts

class POP3Client(basic.LineOnlyReceiver):
    # Indicate whether login() should be allowed if the server
    # offers no authentication challenge and if our transport
    # does not offer any protection via encryption.
    allowInsecureLogin = False

    # Regular expression to match the challenge string in the
    # server greeting line.
    challengeMagicRe = re.compile('<([^>]+)>')

    # Challenge received from the server; set by the default
    # serverGreeting implementation.
    serverChallenge = None

    # List of pending calls.
    # We are a pipelining API but don't actually
    # support pipelining on the network yet.
    _blockedQueue = None

    # Internal helpers, not of much use to users
    def _blocked(self, f, *a):
        if self._blockedQueue is not None:
            d = defer.Deferred()
            self._blockedQueue.append((d, f, a))
            return d
        self._blockedQueue = []
        return None

    def _unblock(self):
        if self._blockedQueue == []:
            self._blockedQueue = None
        elif self._blockedQueue is not None:
            d, f, a = self._blockedQueue.pop(0)
            d2 = f(*a)
            d2.chainDeferred(d)

    def sendShort(self, cmd, args):
        d = self._blocked(self.sendShort, cmd, args)
        if d is not None:
            return d

        if args:
            self.sendLine(cmd + ' ' + args)
        else:
            self.sendLine(cmd)
        self.state = 'SHORT'
        self.waiting = defer.Deferred()
        return self.waiting

    def sendLong(self, cmd, args, consumer, xform):
        d = self._blocked(self.sendLong, cmd, args, consumer)
        if d is not None:
            return d

        if args:
            self.sendLine(cmd + ' ' + args)
        else:
            self.sendLine(cmd)
        self.state = 'LONG_INITIAL'
        self.xform = xform
        self.consumer = consumer
        self.waiting = defer.Deferred()
        return self.waiting

    # Twisted protocol callback
    def connectionMade(self):
        self.state = 'WELCOME'

    def lineReceived(self, line):
        state = self.state
        self.state = None
        state = getattr(self, 'state_' + state)(line) or state
        if self.state is None:
            self.state = state

    def lineLengthExceeded(self, buffer):
        # XXX - We need to be smarter about this
        if self.waiting is not None:
            waiting, self.waiting = self.waiting, None
            waiting.errback(LineTooLong())
        self.transport.loseConnection()

    # POP3 Client state logic - don't touch this.
    def state_WELCOME(self, line):
        code, status = _codeStatusSplit(line)
        if code != OK:
            self.transport.loseConnection()
        else:
            m = self.challengeMagicRe.match(status)
            if m is not None:
                self.serverGreeting(m.group(1))
            else:
                self.serverGreeting(None)
        return 'WAITING'

    def state_WAITING(self, line):
        log.msg("Illegal line from server: " + repr(line))

    def state_SHORT(self, line):
        deferred, self.waiting = self.waiting, None
        self._unblock()
        code, status = _codeStatusSplit(line)
        if code == OK:
            deferred.callback(status)
        else:
            deferred.errback(ServerErrorResponse(status))
        return 'WAITING'

    def state_LONG_INITIAL(self, line):
        code, status = _codeStatusSplit(line)
        if code == OK:
            return 'LONG'
        consumer = self.consumer
        deferred = self.waiting
        self.consumer = self.waiting = self.xform = None
        self._unblock()
        deferred.errback(ServerErrorResponse(status, consumer))
        return 'WAITING'

    def state_LONG(self, line):
        if line == '.':
            consumer = self.consumer
            deferred = self.waiting
            self.consumer = self.waiting = self.xform = None
            self._unblock()
            deferred.callback(consumer)
            return 'WAITING'
        else:
            if self.xform is not None:
                self.consumer(self.xform(line))
            else:
                self.consumer(line)
            return 'LONG'

    # Callbacks - override these
    def serverGreeting(self, challenge):
        self.serverChallenge = challenge

    # External hooks - call these (most of 'em anyway)
    def login(self, username, password):
        """Log into the server.

        If APOP is available it will be used.  Otherwise, if
        the transport being used is SSL, plaintext login will
        proceed.  Otherwise, if the instance attribute
        allowInsecureLogin is set to True, insecure plaintext
        login will proceed.  Otherwise,
        InsecureAuthenticationDisallowed will be raised
        (asynchronously).

        @param username: The username to log in with.
        @param password: The password to log in with.

        @rtype: C{Deferred}
        @return: A deferred which fires when login has
        completed.
        """
        if self.serverChallenge is not None:
            return self._apop(username, password, self.serverChallenge)
        elif self.transport.getHost()[0] == 'SSL' or self.allowInsecureLogin:
            return self._plaintext(username, password)
        else:
            return defer.fail(InsecureAuthenticationDisallowed())

    def _plaintext(self, username, password):
        return self.user(username).addCallback(lambda r: self.password(password))

    def _apop(self, username, password, challenge):
        digest = md5.new(challenge + password).hexdigest()
        return self.apop(username, digest)

    def apop(self, username, digest):
        """Perform APOP login.

        Use the L{login} method instead.
       """
        return self.sendShort('APOP', username + ' ' + digest)

    def user(self, username):
        """Send the user command.

        Use the L{login} method instead.
        """
        return self.sendShort('USER', username)

    def password(self, password):
        """Send the password command.

        Use the L{login} method instead.
        """
        return self.sendShort('PASS', password)

    def delete(self, index):
        """Delete a message from the server.

        @rtype: C{Deferred}
        @return: A deferred which fires when the delete command
        is successful, or fails if the server returns an error.
        """
        return self.sendShort('DELE', str(index + 1))

    def _consumeOrSetItem(self, cmd, args, consumer, xform):
        if consumer is None:
            L = []
            consumer = _ListSetter(L).setitem
            return self.sendLong(cmd, args, consumer, xform).addCallback(lambda r: L)
        return self.sendLong(cmd, args, consumer, xform)

    def _consumeOrAppend(self, cmd, args, consumer, xform):
        if consumer is None:
            L = []
            consumer = L.append
            return self.sendLong(cmd, args, consumer, xform).addCallback(lambda r: L)
        return self.sendLong(cmd, args, consumer, xform)

    def capabilities(self, consumer=None):
        """Retrieve the capabilities supported by this server.

        If L{consumer} is not None, it will be called with each
        capability string as it is received.  Otherwise, the
        returned Deferred will be fired with a list of all the
        capability strings when they have all been received.
        """
        return self._consumeOrAppend('CAPA', None, consumer, None)

    def retrieve(self, index, consumer=None, lines=None):
        """Retrieve a message from the server.

        If L{consumer} is not None, it will be called with
        each line of the message as it is received.  Otherwise,
        the returned Deferred will be fired with a list of all
        the lines when the message has been completely received.
        """
        idx = str(index + 1)
        if lines is None:
            return self._consumeOrAppend('RETR', idx, consumer, None)
        return self._consumeOrAppend('TOP', '%s %d' % (idx, lines), consumer, None)

    def listSize(self, consumer=None):
        """Retrieve a list of the size of all messages on the server.

        If L{consumer} is not None, it will be called with two-tuples
        of message index number and message size as they are received.
        Otherwise, a Deferred which will fire with a list of B{only}
        message sizes will be returned.  For messages which have been
        deleted, None will be be placed in the list.
        """
        return self._consumeOrSetItem('LIST', None, consumer, _listXform)

    def listUID(self, consumer=None):
        """Retrieve a list of the UIDs of all messages on the server.

        If L{consumer} is not None, it will be called with two-tuples
        of message index number and message UID as they are received.
        Otherwise, a Deferred which will fire with of list of B{only}
        message UIDs will be returned.  For messages which have been
        deleted, None will be placed in the list.
        """
        return self._consumeOrSetItem('UIDL', None, consumer, _uidXform)

    def quit(self):
        """Disconnect from the server.
        """
        return self.sendShort('QUIT', None)
