# -*- test-case-name: quotient.test.test_smtp -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import sys
import rfc822
import math
import time
from email.Generator import Generator

from atop import store, sched, credup
from atop.store import transacted
from atop.tpython import Forgetter

from twisted.protocols import smtp
from twisted.python import log
from twisted.python import components
from twisted.internet import error
from twisted.internet import defer
from twisted.persisted import styles
from twisted.names.client import createResolver

from twisted.protocols import smtp
from twisted.mail.relaymanager import MXCalculator

from quotient.copyright import longversion
from quotient.proto import smtpin
from quotient import mimemessage
from quotient import mimebakery
from quotient import provisioner
from quotient import prefup
from quotient.application import MailApp, IMailApp

class ResolverManager:
    """Make a single resolver instance available to anyone who needs one.
    """
    privateResolver = None
    privateCalculator = None
    def makeResolver(self):
        if self.privateResolver is None:
            self.privateResolver = createResolver()
        return self.privateResolver

    def makeCalculator(self):
        if self.privateCalculator is None:
            self.privateCalculator = MXCalculator(self.makeResolver())
        return self.privateCalculator
makeCalculator = ResolverManager().makeCalculator

def msgToFile(msg):
    """This function is Death Made Source Code.  Invoke it at your own peril.

    Return a file which, when read from, returns bytes representing the MIME
    format of the given message.

    Currently, this operates entirely in memory (this is a bad thing).
    """
    s = StringIO()
    Generator(s).flatten(msg)
    s.seek(0, 0)
    return s

def split(addr):
    i = addr.rfind('@')
    if i == -1:
        return addr, ''
    return addr[:i], addr[i+1:]

class IDeliveryInformation(components.Interface):
    def removeAddress(self, local, domain):
        pass
    def deliverySucceeded(self, local, domain):
        pass
    def transmissionFailed(self, local, domain):
        pass
    def forceFailure(self, local, domain):
        pass
    def cleanup(self):
        pass

class DeliveryInformation(store.Item, sched.Schedulable):
    __implements__ = (IDeliveryInformation, sched.Schedulable.__implements__)

    # Exponential backoff tuned to a ~5 day retry period
    RETRANS_TIMES = [40, 400, 4000, 40000, 400000]

    # Test values
    # RETRANS_TIMES = [2, 4, 6, 8]

    # Reference to the outbox
    outbox = store.referenced

    # Reference to the secret, delivery-attempt-management outbox
    deliveryBox = store.referenced

    # Message being sent
    msg = store.referenced

    # Address of sender
    from_ = ''

    # Currently attempting a delivery
    delivering = False

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        try:
            del self.nextAttempt
        except AttributeError:
            pass
        self.reset(0)

    def __init__(self, msg, fromAddr, outbox, deliveryBox, recipients):
        store.Item.__init__(self, msg.store)
        self.msg = msg
        self.from_ = fromAddr
        self.outbox = outbox
        self.deliveryBox = deliveryBox
        # Address -> transmission attempts mapping
        self.recipients = dict.fromkeys(recipients, -1)

        outbox.addItem(msg)
        deliveryBox.addItem(self)
        self.run()

    def run(self):
        log.msg("Beginning delivery attempt of %s" % (self,))
        self.beginDeliveryAttempt(IOutgoingMail(self.msg.store))

    def beginDeliveryAttempt(self, mail):
        self.touch()
        self.delivering = True
        dl = []
        for (local, domain) in self.recipients.keys():
            dl.append(mail.send(local, domain, self))
        return defer.DeferredList(dl)

    def deliverySucceeded(self, local, domain):
        """Indicate delivery to the given recipient has succeeded.

        If this was the last recipient intended to receive this message,
        clean up (destroy) this DeliveryInformation object.
        """
        self.touch()
        self.delivering = False
        self.removeAddress(local, domain)
        if not self.recipients:
            self.cleanup()
    deliverySucceeded = transacted(deliverySucceeded)

    def cleanup(self):
        """Destroy this DeliveryInformation object.
        """
        if self.msg in self.outbox:
            self.outbox.removeItem(self.msg)
        self.delete()
    cleanup = transacted(cleanup)

    def transmissionFailed(self, local, domain):
        """Indicate a delivery attempt to the given recipient has failed.

        If the delivery attempt limit has been exceeded, remove the
        given recipient from the list of those to whom to attempt
        delivery.  Otherwise, schedule another delivery attempt.

        Returns True if this was a terminal failure, false otherwise.
        """
        self.touch()
        self.delivering = False
        x = self.recipients[local, domain] = self.recipients[local, domain] + 1
        if x < len(self.RETRANS_TIMES):
            # Inter-address collisions here don't matter,
            # the next attempt is going to be very similar for
            # all addresses for a particular message
            self.reset(self.RETRANS_TIMES[x])
            return False
        self.forceFailure(local, domain)
        return True
    transmissionFailed = transacted(transmissionFailed)

    def forceFailure(self, local, domain):
        """Indicate terminal delivery failure for the given recipient.

        Remove the recipient from the delivery list.  If they were the
        final recipient, destroy this DeliveryInformation object.
        """
        self.removeAddress(local, domain)
        # XXX Generate a bounce here, maybe?
        if not self.recipients:
            self.cleanup()
    forceFailure = transacted(forceFailure)

    def removeAddress(self, local, domain):
        """Remove the given recipient from the delivery list.
        """
        assert (local, domain) in self.recipients
        self.touch()
        del self.recipients[local, domain]

    def __nonzero__(self):
        return bool(self.recipients)

class IOutgoingMail(components.Interface):
    def sendString(self, messageText, attachments, to, replyTo, fromAddress, cc='', bcc='', subject='(No Subject'):
        pass

def addrlist(s):
    return list(rfc822.AddressList(s))

class OutgoingMail(styles.Versioned):
    __implements__ = (IOutgoingMail,)

    tankID = provisioner.ETank.OUTGOING_MAIL

    sendmail = staticmethod(smtp.sendmail)

    persistenceVersion = 4

    def upgradeToVersion4(self):
        self.store.touch()
        try:
            del self.mx
        except AttributeError:
            pass

    def upgradeToVersion3(self):
        self.store.touch()
        self.delivRef = store.Pool(self.store, "Delivery Tracking Pool").referenceTo()

    def upgradeToVersion2(self):
        from quotient.prefup import IPrefs
        if self.store.name not in ('postmaster', 'janitor', 'admin'):
            ignore = IPrefs(self.store)
            app = IMailApp(self.store)
            app.appRoot.addItem(self.outRef.getItem())
            app.appRoot.addItem(self.sentRef.getItem())
            self.store.getRootPool().removeItem(self.outRef.getItem())
            self.store.getRootPool().removeItem(self.sentRef.getItem())

    def upgradeToVersion1(self):
        from atop.store import IPipe
        from quotient.queue import upgradeIndex
        from quotient.mimemessage import FlagsIndex
        self.avatar.touch()
        t = self.sentRef.getItem()
        t.touch()
        t.__class__ = mimemessage.EmailPool
        t.addIndexes()

    def __init__(self, avatar, outRef, sentRef):
        self.avatar = avatar
        self.outRef = outRef
        self.sentRef = sentRef
        self.delivRef = store.Pool(avatar, "Delivery Tracking Pool").referenceTo()
    store = property(lambda self: self.avatar)

    def sendString(self, messageText, attachments, to, replyTo, fromAddress, cc='', bcc='', subject='(No Subject)', inReplyTo=None):
        """Take some user input and create a message and send it.

        @param messageText: A convenience string for the text/plain part.
        @param to: rfc822-format address list.
        @param fromAddress: rfc822-format address.
        @param cc: rfc822-format address list.
        @param bcc: rfc822-format address list.
        @param subject: str.  the message's subject.
        @param attachments: A list of 3-tuples of (data, mimetype, filename).
        @param inReplyTo: the message-ID string from a message this is in reply to.
        """
        avatar = self.avatar
        to = addrlist(to)
        cc = addrlist(cc)
        bcc = addrlist(bcc)
        sender = list(rfc822.AddressList(fromAddress))[0]
        rec = mimebakery.Recipe(avatar, to, sender=sender, cc=cc or None, bcc=bcc or None, subject=subject, inReplyTo=inReplyTo)
        rec.text(messageText)
        if attachments:
            for data, mimetype, filename in attachments:
                rec.attach(data, mimetype, filename)
        return self.sendMIMEMessage(rec.bake(), fromAddress, to + cc + bcc)

    def sendMIMEMessage(self, msg, fromAddr, toAddrs):
        return DeliveryInformation(msg, fromAddr,
                                   self.outRef.getItem(),
                                   self.delivRef.getItem(),
                                   [split(x[1]) for x in toAddrs])

    def getCalculator(self):
        return makeCalculator()

    def send(self, localpart, domain, dInfo):
        """
        (internal)

        @param from_: The address from which to send this message.
        @param localpart: The local part of the recipient of this message.
        @param domain: The domain part of the recipient of this message.
        @param msg: A MIMEMessage to send to the specified recipient.
        """
        d = self.getCalculator().getMX(domain)
        d.addCallback(self._cbGotExchange, localpart, domain, dInfo)
        d.addCallback(self._cbMailsent, localpart, domain, dInfo)
        d.addErrback(self._ebGotExchange, localpart, domain, dInfo)
        d.addErrback(log.err)
        return d

    def _cbGotExchange(self, mx, l, domain, dInfo):
        from_ = dInfo.from_
        msgFile = msgToFile(dInfo.msg)
        return self.sendmail(str(mx.exchange), from_, ['@'.join((l, domain))], msgFile)

    def _cbMailsent(self, result, localpart, domain, dInfo):
        dInfo.deliverySucceeded(localpart, domain)
        sent = self.sentRef.getItem()
        if dInfo.msg not in sent:
            sent.addItem(dInfo.msg)
        return result
    _cbMailsent = transacted(_cbMailsent)

    def _ebGotExchange(self, failure, localpart, domain, dInfo):
        from_ = dInfo.from_
        # Check the retry count on this message
        if dInfo.transmissionFailed(localpart, domain):
            log.msg("Transmission failed, bouncing message "
                    "destined for %s@%s back to %s" % (localpart,
                                                       domain, from_))
            if failure.check(smtp.SMTPDeliveryError):
                return self.generateBounce(from_, localpart, domain, dInfo.msg, failure.value.log)
            return self.generateBounce(from_, localpart, domain, dInfo.msg, '')
        elif failure.check(smtp.SMTPDeliveryError):
            addresses = failure.value.addresses
            assert len(addresses) == 1, "Too many responses! %s" % (addresses,)
            code = addresses[0][1]
            if 500 <= code < 600:
                log.msg("Fatal delivery error (SMTP code %d), "
                        "bouncing message destined for %s@%s "
                        "back to %s" % (code, localpart, domain, from_))
                # Permanent failure, bounce it.
                dInfo.forceFailure(localpart, domain)
                return self.generateBounce(from_, localpart, domain, dInfo.msg, failure.value.log)
            else:
                log.msg("Transient delivery error "
                        "(SMTP code %d), rescheduling "
                        "delivery to %s@%s for later." % (code, localpart, domain))
                return self._transient(dInfo)
        elif failure.check(error.DNSLookupError):
            log.msg("DNS lookup error (%s), rescheduling "
                    "delivery to %s@%s for later." % (failure.value,
                                                      localpart, domain))
            return self._transient(dInfo)
        elif failure.check(defer.TimeoutError, error.TimeoutError):
            log.msg("Some kind of timeout error (%s), rescheduling "
                    "delivery to %s@%s for later." % (failure.value,
                                                      localpart, domain))
            return self._transient(dInfo)
        elif failure.check(error.ConnectionRefusedError):
            log.msg("Connection refused error (%s), rescheduling "
                    "delivery to %s@%s for later." % (failure.value,
                                                      localpart, domain))
            return self._transient(dInfo)
        else:
            # Permanent failure
            log.msg("Unknown error, bouncing message "
                    "destined for %s@%s back to %s" % (localpart, domain, from_))
            log.err(failure)
            dInfo.forceFailure(localpart, domain)
            xscript = failure.getTraceback()
            return self.generateBounce(from_, localpart, domain, dInfo.msg, xscript)
    _ebGotExchange = transacted(_ebGotExchange)

    def _transient(self, dInfo):
        try:
            fmt = "Transient delivery error.  Rescheduled %s for delivery at %s."
            args = (dInfo.msg.getItemPath(), time.ctime(dInfo.scheduled))
            log.msg(fmt % args)
        except:
            log.err()

    def generateBounce(self, ffrom, localpart, domain, msg, transcript):
        from quotient.mimebakery import Recipe
        fto = '@'.join((localpart, domain))
        posty = '@'.join(('postmaster', domain))
        subj = "We were unable to deliver your message to %s." % fto
        bounceMsg = Recipe(self.store, to=[('',ffrom)], sender=("Postmaster", posty), subject=subj)
        bText = 'Your message to %s, subject "%s" could not\n be delivered.\n\n' % (fto, msg['subject'])
        bText += 'The text of the message follows:\n\n'
        bounceMsg.text(bText + '\n' + msg.getDisplayPart().get_payload())
        if transcript:
            bounceMsg.attach(transcript, mimetype='text/plain', filename='exception.txt')
        mr = mimemessage.IMIMEDelivery(self.avatar).createMIMEReceiver(True)
        return mr.feedFile(msgToFile(bounceMsg.bake())).addErrback(self._ebBounce, msg)

    def _ebBounce(self, failure, msg):
        log.err(failure)
        log.msg("Bounce generation failed!  Putting message back in outbox.")
        self.outRef.getItem().addItem(msg)

_internalServerError = smtp.SMTPServerError(550, "Internal server error")

class Message:
    """A message being delivered to the ESMTP server for relaying."""

    __implements__ = smtp.IMessage,

    def __init__(self, avatar, sender):
        self.avatar = avatar
        m = mimemessage.MIMEMessageReceiver(avatar, sender, False)
        m.makeConnection(None)
        self.receiver = m

    def lineReceived(self, line):
        return self.receiver.lineReceived(line)

    def eomReceived(self):
        try:
            self.receiver.messageDone()
        except:
            log.err()
            return defer.fail(_internalServerError)
        return defer.succeed("message delivered")

class TeeMessage:
    def __init__(self, one, two):
        self.one = one
        self.two = two

    def lineReceived(self, line):
        self.one.lineReceived(line)
        self.two.lineReceived(line)

    def eomReceived(self):
        return defer.DeferredList([
            defer.maybeDeferred(m.eomReceived) for m in [self.one, self.two]
            ], consumeErrors=True).addCallback(self._cbEOM)
    
    def _cbEOM(self, resultList):
        for (success, result) in resultList:
            if not success:
                return defer.fail(_internalServerError)
    
class MessageDelivery:
    __implements__ = (smtp.IMessageDelivery,)

    def __init__(self, avatar):
        self.store = avatar
        omail = IOutgoingMail(self.store)
        self.senderDelivery = lambda: Message(self.store, omail.sentRef.getItem().addItem)

    def getDomain(self):
        def _():
            return credup.ICredentialsChecker(self.store.store).getDomains().next()
        return self.store.transact(_).name

    def getDomains(self):
        def _():
            return [d.name for d in credup.ICredentialsChecker(self.store.store).getDomains()]
        return self.store.transact(_)

    def receivedHeader(self, helo, origin, recipients):
        # XXX violate that abstraction!  *violate* it!
        domain = self.getDomain()
        heloStr = ""
        if helo[0]:
            heloStr = " helo=%s" % (helo[0],)
        auth = " AUTH=" + self.store.name.encode("xtext")
        from_ = "from %s ([%s]%s)%s" % (helo[0], helo[1], heloStr, auth)
        by = "by %s with ESMTP (%s)" % (domain, longversion)
        for_ = "for %s; %s" % (" ".join(map(str, recipients)), smtp.rfc822date())
        return "Received: %s\n\t%s\n\t%s" % (from_, by, for_)

    def validateFrom(self, helo, origin):
        # The only from address we accept is that of the authenticated user
        if origin.local != self.store.name:
            raise smtp.SMTPBadSender(origin)
        # XXX - This isn't a particularly efficient way to check
        if origin.domain not in self.getDomains():
            raise smtp.SMTPBadSender(origin)
        return origin

    def getPortal(self):
        # XXX XXX XXX XXX XXX
        # This is so messed up and wrong and stupid and sucky
        # XXX XXX XXX XXX XXX
        return self.store.store.service.parent.portal

    def validateTo(self, user):
        omail = IOutgoingMail(self.store)

        receiverDelivery = None
        for d in self.getDomains():
            if user.dest.domain == '' or user.dest.domain == d:
                receiverDelivery = smtpin.validateTo(self.getPortal(), user)
                break
        else:
            f = self.store.name + '@' + self.getDomain()
            l = user.dest.local
            d = user.dest.domain
            sender = lambda m: omail.sendMIMEMessage(m, f, addrlist(l + '@' + d))
            receiverDelivery = defer.succeed(lambda: Message(self.store, sender))

        # We want a copy of this message for our Sent box
        if self.senderDelivery is not None:
            senderDelivery = self.senderDelivery
            self.senderDelivery = None
            return receiverDelivery.addCallback(
                lambda rece: lambda: TeeMessage(senderDelivery(), rece()))
        else:
            return receiverDelivery


class OutgoingMailup(provisioner.ProvisionedPowerup):
    __implements__ = (provisioner.ProvisionedPowerup.__implements__,
                      provisioner.IProvisionable,
                      sched.IScheduledEvent,
                      smtp.IMessageDeliveryFactory)

    dependencies = [provisioner.Provisioner, sched.SubScheduler, prefup.Prefup, MailApp]
    storeDependencies = [sched.SystemScheduler]

    powerupName = 'Compose'

    interval = 30

    startCost = 1
    tankID = provisioner.ETank.OUTGOING_MAIL

    persistenceVersion = 2

    def upgradeToVersion2(self):
        self.touch()
        from atop.upgrade import versionedUpgrade
        versionedUpgrade(IOutgoingMail(self.store).delivRef.getItem())

    def upgradeToVersion1(self):
        self.touch()
        self.store.unsetComponent(smtp.IMessageDelivery)
        self.store.setComponent(smtp.IMessageDeliveryFactory, self)

    def setUpPools(self, avatar):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)

        app = IMailApp(avatar)
        outbox = store.Pool(avatar, name="Outbox")
        sent = mimemessage.EmailPool(avatar, name="Sent")

        app.appRoot.addItem(outbox)
        app.appRoot.addItem(sent)

        avatar.setComponent(IOutgoingMail,
                            OutgoingMail(avatar,
                                         outbox.referenceTo(),
                                         sent.referenceTo()))
        avatar.setComponent(smtp.IMessageDeliveryFactory, self)

    def getMessageDelivery(self):
        return MessageDelivery(self.store)

    def run(self):
        pass
