import struct, socket
from twisted.internet import protocol
from twisted.python import log
from twisted.application import service, internet
MAPPED_ADDRESS     = 0x0001
RESPONSE_ADDRESS   = 0x0002
CHANGE_REQUEST     = 0x0003
SOURCE_ADDRESS     = 0x0004
CHANGED_ADDRESS    = 0x0005
USERNAME           = 0x0006
PASSWORD           = 0x0007
MESSAGE_INTEGRITY  = 0x0008
ERROR_CODE         = 0x0009
UNKNOWN_ATTRIBUTES = 0x000a
REFLECTED_FROM     = 0x000b

BINDING_REQUEST               = 0x0001 
BINDING_RESPONSE              = 0x0101 
BINDING_ERROR_RESPONSE        = 0x0111 
SHARED_SECRET_REQUEST         = 0x0002 
SHARED_SECRET_RESPONSE        = 0x0102 
SHARED_SECRET_ERROR_RESPONSE  = 0x0112 


class STUNServer(protocol.DatagramProtocol):
    chatty = 0
    def chat(self, msg):
        if self.chatty:
            log.msg(msg)
    def datagramReceived(self, dgram, addr, secondary=0):
        clientHost, clientPort = addr
        if secondary:            
            dt = self.alttransport
            ct = self.transport
            dp = self.service.altport
            cp = self.service.port
        else:
            dt = self.transport
            ct = self.alttransport
            dp = self.service.port
            cp = self.service.altport
        self.chat("!!! Arrived on %s" % dp)
        msgType, msgLength, tid = struct.unpack("!hh16s", dgram[:20])
        i = 20
        requestAttrs = {}
        extensions = {}

        while i < msgLength+20:
            attrType, attrLength = struct.unpack("!hh", dgram[i:i+4])
            i += 4
            attrValue = dgram[i:i+attrLength]
            i += attrLength
            if attrType > 0x7fff:
                extensions[attrType] = attrValue
            elif attrType > 0x000c:
                return self.errorResponse(addr, 420, "Unsupported attribute 0x%04x" % attrType, {UNKNOWN_ATTRIBUTES: attrType})
            requestAttrs[attrType] = attrValue
        self.chat("STUN packet from %s received: attrs %s" % (addr,requestAttrs))
        if (msgType == BINDING_RESPONSE) and addr[0] == self.service.althost:
            #if this response has been forwarded to us by our coserver, send it on
            dest = requestAttrs.get(REFLECTED_FROM, requestAttrs[MAPPED_ADDRESS])
            destPort, destHost = struct.unpack("!xxh",dest[:4])[0], socket.inet_ntoa(dest[4:8])
            self.chat("Forwarding response created by coserver to %s, %s via %s" % (destPort, destHost, clientPort))
            dt.write(dgram, (destHost, destPort))            
            return
        
        elif msgType != BINDING_REQUEST:
            return self.errorResponse(400, "Unsupported message type 0x%04x" % msgType)

        if requestAttrs.has_key(CHANGE_REQUEST):
            self.chat("Change request received.")
            change = struct.unpack("!i", requestAttrs[CHANGE_REQUEST])[0]
            changeIP, changePort = change & 0x4, change & 0x2            
        else:
            changeIP, changePort = 0, 0

        if changePort:
            p = cp
            self.chat("Changing port.")
        else:
            p = dp
            
        if changeIP:
            h = self.service.althost
            self.chat("Changing host.")
        else:
            h = self.service.host
            
        responseAttrs = {}
        responseAttrs[MAPPED_ADDRESS] =  struct.pack("!xbh", 0x01, clientPort) + socket.inet_aton(clientHost)
        responseAttrs[CHANGED_ADDRESS] = struct.pack("!xbh", 0x01, cp) + socket.inet_aton(self.service.althost)
        responseAttrs[SOURCE_ADDRESS] =  struct.pack("!xbh", 0x01, p) + socket.inet_aton(h)
        if requestAttrs.has_key(RESPONSE_ADDRESS):
            responseAttrs[REFLECTED_FROM] = struct.pack("!xbh", 0x01, clientPort) + socket.inet_aton(clientHost)
            destPort, destHost = struct.unpack("!xxh", requestAttrs[RESPONSE_ADDRESS][:4])[0], socket.inet_ntoa(requestAttrs[RESPONSE_ADDRESS][4:8])
        else:
            destHost, destPort = clientHost, clientPort
            
        responseAttrs.update(extensions)
        
        attrString = "".join([struct.pack("!hh", k, len(v)) + v for k,v in responseAttrs.items()])
        header = struct.pack("!hh16s", BINDING_RESPONSE, len(attrString), tid)
        message = header + attrString

        #now, where do we send this

        if changeIP:
            destHost = self.service.althost
            destPort = p            
        if changePort:            
            ct.write(message, (destHost, destPort))
            self.chat("!!! Sent on %s" % ct.socket.getsockname()[1])
        else:
            dt.write(message, (destHost, destPort))
            self.chat("!!! Sent on %s" % dt.socket.getsockname()[1])
            
class SecondarySTUNServer(protocol.DatagramProtocol):
    
    def __init__(self, original):
        self.original = original

    def datagramReceived(self, dgram, addr):
        self.original.datagramReceived(dgram, addr, secondary=1)

class STUNService(internet.UDPServer):

    def __init__(self, host, port, althost, altport, proto):
        self.host = host
        self.port = port
        self.althost = althost
        self.altport = altport
        self.proto = proto
        self.proto.service = self
    def _getPort(self):
        from twisted.internet import reactor
        self._altConnection = reactor.listenUDP(self.altport, SecondarySTUNServer(self.proto), interface=self.host)
        self.proto.alttransport = self._altConnection
        return reactor.listenUDP(self.port, self.proto, interface=self.host)
