# -*- test-case-name: quotient.test.test_extract -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from atop import store
from atop import tpython
from atop import powerup
from atop import upgrade

from nevow import tags

from quotient import cleanup
from quotient import provisioner
from quotient import mimemessage
from quotient import taglib
from quotient.web.root import Base
from quotient.prefup import IPrefs
from quotient.application import MailApp, IMailApp

from twisted.python.components import Interface

class FromMessageIndex:
    __implements__ = (store.IPoolIndex,)

    def getIndexName(self):
        return "fmsg"

    def keyToSortableString(self, key):
        return '%010d' % (key,)

    def sortableStringToKey(self, sstr):
        return int(sstr)


class IExtractable(Interface):
    pass

class ISmartenup(Interface):
    def smartsPool(self):
        """Return the pool containing extraction smarts items"""

    def subscribeItem(self, smart):
        """Subscribe the given extraction smarts item to the item added
        event notification list.
        """

    def subscribePart(self, smart, type='text/plain', decode=True):
        """Subscribe the given extraction smarts item to the notification
        list for message parts which match the given type.
        """

class IInbox(Interface):
    inbox = property(doc="The inbox pool")

class Smartenup(provisioner.ProvisionedPowerup):

    __implements__ = (provisioner.ProvisionedPowerup.__implements__,
                      provisioner.IProvisionable, ISmartenup, IInbox)

    dependencies = [provisioner.Provisioner, cleanup.Cleanup,  MailApp]
    invisible = True
    totalExtracted = 0
    tankID = provisioner.ETank.EXTRACTION
    startCost = 1
    byteCost = 0.0001
    powerupName = "Smartenup"

    inbox = store.referenced

    persistenceVersion = 3

    def upgradeToVersion3(self):
        self.touch()
        self.inbox.setIndexKey('name', 'Filtered')

    def upgradeToVersion2(self):
        self.touch()
        self.inbox = self.inboxRef.getItem()
        del self.inboxRef

    def upgradeToVersion1(self):
        # TODO - Get rid of notey the Notebook ref
        self.touch()
        from quotient.prefup import IPrefs
        ignore = IPrefs(self.store)
        app = IMailApp(self.store)
        app.appRoot.addItem(self.inboxRef.getItem())
        self.store.getRootPool().removeItem(self.inboxRef.getItem())

    def setUpPools(self, avatar, name=None):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)

        cleaner = cleanup.IClean(avatar)
        cleaner.cleanPool.addPipe(self.fwd(store.IPipe))

        avatar.setComponent(ISmartenup, self)
        avatar.setComponent(IInbox, self)

        smartsPool = store.Pool(avatar, "Extracts")
        smartsPool.addIndex(store.NameIndex())
        self.smartsPoolRef = smartsPool.getRef()
        avatar.getRootPool().addItem(smartsPool)

        # dict of smartRef to type string
        self.partSmarts = {}
        # list of smartRefs that want the whole item
        self.itemSmarts = []

        # Arg special case arg
        inbox = mimemessage.EmailPool(avatar, 'Filtered') # was Inbox!!!
        self.inbox = inbox
        app = IMailApp(avatar)
        app.appRoot.addItem(inbox)

    def smartsPool(self):
        return self.smartsPoolRef.getItem()

    def subscribeItem(self, smart):
        self.touch()
        self.smartsPool().addItem(smart.extractPool())
        self.itemSmarts.append(smart.referenceTo())

    def subscribePart(self, smart, type='text/plain', decode=True):
        self.touch()
        self.smartsPool().addItem(smart.extractPool())
        self.partSmarts[smart.referenceTo()] = (type, decode)

    def itemAdded(self, item):
        # QA METRICS
        # to short-circuit extraction, uncomment the following line:
        #return
        if not self.extractAndFile(item):
            self.inbox.addItem(item)

    def extractAndFile(self, item, itemSmarts=None, partSmarts=None):
        if itemSmarts is None:
            itemSmarts = self.itemSmarts
        if partSmarts is None:
            partSmarts = self.partSmarts

        if isinstance(item, mimemessage.MIMEMessage):
            return self.runFilers(item, itemSmarts, partSmarts)
        else:
            self.runPartExtractors(item, partSmarts)
            return False

    def runPartExtractors(self, partSmarts):
        for smartRef, (type, decode) in partSmarts.iteritems():
            if type.startswith('text/plain'):
                smart = smartRef.getItem()
                eItem = IExtractable(item, default=None)
                if eItem is not None:
                    smart.extractPart(item, item.description, None)

    def runFilers(self, item, itemSmarts, partSmarts):
        filed = False
        for smartRef in itemSmarts:
            if smartRef.getItem().extractItem(item) == "filed":
                filed = True
        self.extractItem(item, partSmarts)
        return filed

    def itemRemoved(self, item):
        if isinstance(item, mimemessage.MIMEMessage):
            smarts = self.partSmarts.keys() + self.itemSmarts
            for smartRef in smarts:
                smartRef.getItem().itemRemoved(item)
            self.removeFromInbox(item)

    def removeFromInbox(self, item):
        if item in self.inbox:
            self.inbox.removeItem(item)

    def itemChanged(self, item, changeData):
        allSmarts = self.itemSmarts + self.partSmarts.keys()
        for smartRef in allSmarts:
            smart = smartRef.getItem()
            smart.itemChanged(item, changeData)

    def extractLoop(self, msg, smarts):
        avatar = self.store
        provis = provisioner.IProvisioner(avatar)
        for part in msg.walk():
            pType = part.inferType()
            for smartRef, (type, decode) in smarts.iteritems():
                if pType.startswith(type):
                    smart = smartRef.getItem()
                    payload = part.get_payload(decode)
                    cost = len(payload) * smart.byteCost
                    if True: #provis.deplete(smart.tankID, cost):
                        smart.extractPart(msg, payload, part)
                        if len(getattr(msg, 'extracts', ())) > 30:
                            return
                        yield None

    def extractItem(self, msg, smarts):
        def _(i):
            for e in i:
                pass
        self.store.transact(_, self.extractLoop(msg, smarts))

class SmartenupDisplay(Base):
    document = taglib.page(title="Smartenup")[
        tags.p["Hello"],
    ]

class ReExamineItems(store.Item):
    smarts = store.referenced

    def __init__(self, st, smarts, itemSmarts=None, partSmarts=None):
        store.Item.__init__(self, st)
        self.smarts = smarts
        self.itemSmarts = itemSmarts
        self.partSmarts = partSmarts

    def __call__(self, pool, item):
        if self.smarts.extractAndFile(item, self.itemSmarts, self.partSmarts):
            if item in self.smarts.inbox:
                inbox.removeItem(item)

def upgradeSmarts(avatar, itemSmarts=None, partSmarts=None):
    clean = cleanup.IClean(avatar).cleanPool
    smarts = ISmartenup(avatar)
    examiner = ReExamineItems(avatar, smarts, itemSmarts, partSmarts)
    return upgrade.ItemUpgradeFactoryUpgradeQueue(avatar, clean, examiner).start()

def retroactivelyExtract(avatar, PowerupType):
    """Install an Extract Powerup and apply it retroactively.

    All messages in the Clean pool are passed to the new extractor
    and, if claimed, removed from the inbox.

    The instantiated Powerup is returned in a two-tuple, along with
    a Deferred that fires when the retroactive application has finished.
    """
    ps = powerup.IPowerStation(avatar)
    pup = ps.installPowerup(PowerupType)

    # XXX - Some duplication with ExtractCollectorBase.subscribe
    if pup.type is None:
        d = upgradeSmarts(avatar, itemSmarts=[pup.referenceTo()])
    else:
        d = upgradeSmarts(avatar, partSmarts={pup.referenceTo(): (pup.type, pup.decode)})
    return pup, d
