# -*- test-case-name: quotient.test.test_extract -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time

from atop.store import Item, Pool, ItemReferenceIndex, IPoolIndex, referenced
from atop import powerup

from nevow import tags

from quotient import provisioner
from quotient.mimemessage import MIMEMessage
from quotient import taglib
from quotient.web.root import Base
from quotient.smartenup import Smartenup, ISmartenup
from quotient.prefup import IPrefs

from twisted.python.components import Interface

# extractItem returns one of these, or None.
# This one means it put the Item somewhere else, but
# it is alright to put it in the Inbox
CLASSIFIED = 'classified'
# This one means it created a new Item
EXTRACTED = 'extracted'
# This one means it put it somewhere and it should not
# end up in the Inbox
FILED = 'filed'

class ParentIndex:
    __implements__ = (IPoolIndex,)

    def getIndexName(self):
        return "parent"

    def keyToSortableString(self, key):
        return '%10d' % (key,)

    def sortableStringToKey(self, sstr):
        return int(sstr)

class PrimaryKeyIndex:
    __implements__ = (IPoolIndex,)

    def getIndexName(self):
        return 'pkey'

    def keyToSortableString(self, key):
        return key

    def sortableStringToKey(self, sstr):
        return sstr

class ExtractCollectorBase(provisioner.ProvisionedPowerup):

    __implements__ = provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable

    dependencies = [provisioner.Provisioner, Smartenup]
    invisible = True
    totalExtracted = 0
    tankID = provisioner.ETank.EXTRACTION
    startCost = 1
    byteCost = 0.0001

    maxExtractsPer = 30

    poolOfExtracts = referenced

    def setUpPools(self, store, name=None, type='text/plain', decode=True):
        provisioner.ProvisionedPowerup.setUpPools(self, store)
        if name is None:
            name = self.powerupName
        self.setupExtractPool(name)
        self.type = type
        self.decode = decode
        self.subscribe(type, decode)

    def setupExtractPool(self, name):
        self.poolOfExtracts = p = Pool(self.store, name=name)
        p.addIndex(PrimaryKeyIndex())
        p.suggestIndex('received', backwards=True)
        p.addIndex(ItemReferenceIndex('creator', self.store))

    def subscribe(self, type, decode=True):
        smarts = ISmartenup(self.store)
        if type is None:
            smarts.subscribeItem(self)
        else:
            smarts.subscribePart(self, type, decode)

    def extractPool(self):
        return self.poolOfExtracts

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, changeData):
        pass

    def foundExtract(self, item, extract):
        self.touch()
        item.touch()
        self.totalExtracted += 1
        if not hasattr(item, 'extracts'):
            item.extracts = []
        extract.setIndexKey('creator', extract.creatorRef)
        item.extracts.append(extract.getRef())

    def createExtract(self, pkey, *args, **kw):
        p = self.extractPool()
        i = p.getIndexItem('pkey', pkey, default=None)
        if i is None:
            i = self.ExtractType(self.store, *args, **kw)
            i.setIndexKey('pkey', pkey)
            p.addItem(i)
        return i

    def queryExtracts(self, item):
        """Return an iterator of Extracts of this type for the given message.
        """
        p = self.extractPool()
        key = item.storeID
        return p.queryIndex('parent', startKey=key, endKey=key)

    def extractsFrom(self, person):
        """Return an iterator of Extracts of this type for a give Person.
        """
        p = self.extractPool()
        key = person.getRef()
        return p.queryIndex('creator', startKey=key, endKey=key)


class ExtractDetail(Base):
    def render_parent(self, context, data):
        m = data.itemRef.getItem()
        if isinstance(m, MIMEMessage):
            subj = m['subject'] or '<No Subject>'
            frm = m['from'] or '<No Sender>'
        else:
            subj = m.description
            frm = m.creatorRef.getItem().name
        return [tags.div[taglib.href(m)[taglib.headerToUTF8(subj)]],
                tags.div[[taglib.headerToUTF8(frm)]],]

    def render_context(self, context, extract):
        pad = 50
        text, begin, end = extract.excerpt(50)
        return [
            tags.span(_class="normal")[ text[0:begin] ],
            tags.span(_class="highlight") [ text[begin:end] ],
            tags.span(_class="normal") [ text[end:] ],
        ]

    document = taglib.page(title="Extract")[
        tags.p["Found in: ", render_parent],
        tags.p[render_context],
    ]

class IExtract(Interface):
    pass

class ExtractBase(Item):
    __implements__ = IExtract

    pkey = None

    # Indicates the user has previously told us this extract
    # is not worth dealing with.
    uninteresting = False

    def __init__(self, store, item, description='Extract', part=None):
        Item.__init__(self, store, description)
        self.description = description
        self.itemRef = item.referenceTo()
        if part:
            self.partNum = part.mimeID
        else:
            self.partNum = None
        self.date = time.time()
        self.creatorRef = item.creatorRef

    def setindex_pkey(self, pkey):
        self.pkey = pkey
        return pkey

    def index_pkey(self, pool):
        return self.pkey

    def excerpt(self, pad=None):
        return (self.description, 0, len(self.description))

##    def detach(self, store):
##        from copy import copy
##        dup = copy(self)
##        Item.__init__(dup, store, dup.description)
##        return dup

    def toStan(self):
        return self.description


class Textract(ExtractBase):
    __implements__ = IExtract

    defaultPad = 25

    def __init__(self, store, item, text, begin, end, part=None):
        self.begin = begin
        self.end = end
        self.description, self.descBegin, self.descEnd = self.excerpt(self.defaultPad, text)
        ExtractBase.__init__(self, store, item, self.description, part)
        if part is None:
            self.text = text
            self.partNum = None
        else:
            self.partNum = part.mimeID

    def getText(self, decode=True):
        if self.partNum is None:
            return self.text
        else:
            item = self.itemRef.getItem()
            part = item.getPartByID(self.partNum)
            return part.get_payload(decode)

    def excerpt(self, pad=None, text=None):
        if text is None:
            text = self.getText()
        if pad == None or  pad + pad > len(text):
            return (text, self.begin, self.end)
        offset = max(0, self.begin - pad)
        finish = min(len(text), self.end + pad)
        return (text[offset:finish], self.begin - offset, self.end - offset)

    def toStan(self):
        ### TODO: Temporary expedient hack.
        from nevow.tags import a
        D, B, E = self.description, self.descBegin, self.descEnd
        return D[:B], a(href=self.getLink())[ D[B:E] ], D[E:]

    def getLink(self):
        ### TODO: Temporary expedient hack.
        return self.description[self.descBegin:self.descEnd]
