# -*- test-case-name: quotient.test.test_listup -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import re

from atop import store
from atop import powerup

from twisted.python import components

from quotient import mimemessage
from quotient.smarts import base
from quotient.prefup import IPrefs
from quotient.application import MailApp, IMailApp

class Filer(base.ExtractCollectorBase):
    pass

def matchEcartis(item):
    sender = item['sender']
    xlist = item['x-list']

    if sender and xlist:
        domain = sender.rfind('@')
        if domain != -1 and item['x-ecartis-version']:
            return xlist + '.' + sender[domain + 1:]
    return None

def matchMailman(item, listExpr = re.compile(r'''<(?P<listId>[^>]*)>'''), versionHeader = 'X-Mailman-Version'):
    if versionHeader not in item:
        return None
    listId = item['List-Id']
    if listId:
        m = listExpr.search(listId)
        if m:
            return m.group('listId')
        return listId
    return None

def matchMajorDomo(item, listExpr = re.compile(r'''<(?P<listId>[^>]*)>''')):
    listId = item['X-Mailing-List']
    if listId:
        m = listExpr.search(listId)
        if m:
            return m.group('listId').replace('@', '.')
        return listId.replace('@', '.')
    return None

def matchLyris(item):
    msgid, unsub = item['Message-ID'], item['List-Unsubscribe']
    if msgid and "LYRIS" in msgid and unsub:
        m = re.search(r"""<mailto:leave-(?P<listname>.*)-.*@(?P<domain>.*)>""", unsub)
        if m:
            return "%s.%s" % (m.group('listname'), m.group('domain'))
    return None

def matchRequestTracker(item):
    ticket, managedby = item['RT-Ticket'], item['Managed-by']
    if ticket and managedby:
        if managedby.startswith("RT"):
            return "request-tracker"
        else:
            return None
    else:
        return None

class MailingListFiler(Filer):

    powerupName = "Mailing List Filer"

    persistenceVersion = 2

    matchers = [matchMailman, matchMajorDomo, matchEcartis, matchLyris, matchRequestTracker]
    dependencies = [MailApp]

    def upgradeToVersion1(self):
        from quotient.queue import upgradeIndex
        self.touch()
        p = self.extractPool()
        p.touch()
        p.addIndex(store.NameIndex('imapname'))
        upgradeIndex(p, 'imapname')

    def upgradeToVersion2(self):
        from quotient.prefup import IPrefs
        ignore = IPrefs(self.store)
        app = IMailApp(self.store)
        self.touch()
        app.appRoot.addItem(self.extractPool())
        self.store.getRootPool().removeItem(self.extractPool())

    def setUpPools(self, avatar):
        Filer.setUpPools(self, avatar, "Mailing Lists", None, False)
        self.extractPool().addIndex(store.NameIndex())
        self.extractPool().addIndex(store.NameIndex('imapname'))
        app = IMailApp(avatar)
        app.appRoot.addItem(self.extractPool())

    def extractItem(self, item):
        listName = self.listNameFromItem(item)
        if listName:
            lists = self.extractPool()
            listPool = lists.getIndexItem('name', listName, None)
            if listPool is None:
                listPool = mimemessage.EmailPool(self.store, name=listName)
                lists.addItem(listPool)
            if item not in listPool:
                listPool.addItem(item)
            return base.FILED
        return None

    def listNameFromItem(self, item):
        for f in self.matchers:
            n = f(item)
            if n is not None:
                return n
        return None

class IRouteTarget(components.Interface):
    def receive(self, route, item):
        pass

class Router(Filer):
    greyedOut = True
    invisible = True

    def setUpPools(self, avatar):
        Filer.setUpPools(self, avatar, "Recipient Routing", None, False)

    def extractItem(self, item):
        route = getattr(item, "route", None)
        if route:
            return self.deliverRouted(route, item)
        return None

    def deliverRouted(self, route, item):
        powerupName = route.pop()
        p = powerup.IPowerStation(self.store).getPowerupsPool()
        for pup in p.queryIndex():
            if pup.powerupName.lower() == powerupName.lower():
                deliver = IRouteTarget(pup, default=None)
                if deliver is not None:
                    if deliver.receive(route, item):
                        return base.FILED
        return None
