
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import sys

from twisted.test import proto_helpers
from twisted.protocols import smtp
from twisted.internet import defer
from twisted.protocols.dns import Record_MX
from twisted.python import failure, log
from twisted.internet import reactor
from twisted.cred import portal, checkers

from atop import credup
from atop import powerup
from atop import sched
from atop.store import ITEM
from atop.test import unittest

from quotient.test import emaildata, helper

from quotient import sysup
from quotient import provisioner
from quotient.proto import smtpin
from quotient.proto import smtpout
from quotient.mimemessage import IMIMEDelivery, MIMEMessageReceiver, EmailPowerup

from twisted.cred.portal import Portal, IRealm

class Delivery(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [credup.StoreRealm]

    def dbSetUp(self):
        self.checker = credup.ICredentialsChecker(self.store)
        self.avatar = self.createAvatar([EmailPowerup], "USAR")

        self.checker.addDomain("DOM")
        self.checker.addAvatar(self.avatar, "DOM")
        p = Portal(IRealm(self.store))
        p.registerChecker(self.checker)
        self.delivery = smtpin.QuotientMessageDelivery(p)

    def u(self, s):
        return smtp.User(smtp.Address(s), "HELO", None, smtp.Address("foo@bar"))

    def testValidateFrom(self):
        origin = "this is a bogus string.  we currently accept all origins."
        self.assertEquals(self.delivery.validateFrom("EHLO FROM ME", origin), origin)

    def testInvalidValidateTo(self):
        unittest.deferredError(self.delivery.validateTo(self.u("bogus!"))).trap(
            smtp.SMTPBadRcpt)
        unittest.deferredError(self.delivery.validateTo(self.u("user@domain"))).trap(
            smtp.SMTPBadRcpt)

    def dbtestExceededQuota(self):
        cp = sysup.IControlPanel(self.store)
        cp.allowAction(sysup.IControlPanel.PROVISIONING, True)
        for i in [10000, 100, 1]:
            try:
                while True:
                    provisioner.IProvisioner(self.avatar).deplete(provisioner.ETank.QUOTA, i)
            except provisioner.NotEnoughFuel:
                pass
        self.assertRaises(smtp.SMTPBadRcpt, self.delivery.validateTo, self.u("USAR@DOM"))

    dbtestExceededQuota.skip = "Quotas don't really work right now."

    def dbtestSuccessfulDelivery(self):
        md = unittest.deferredResult(self.delivery.validateTo(self.u("USAR@DOM")))
        msg = md()
        for line in emaildata.test_messages[0].splitlines():
            msg.lineReceived(line)
        d = unittest.deferredResult(msg.eomReceived())
        self.failUnless(d)

        p = IMIMEDelivery(self.avatar).arrivalRef.getItem()
        msgs = self.store.transact(list, p.queryIndex(result=ITEM))
        self.assertEquals(len(msgs), 1)
        self.assertEquals(msgs[0].openFile().read(), emaildata.test_messages[0])

class MXHijacker:
    def __init__(self, hosts):
        self.hosts = hosts

    def getMX(self, dname):
        try:
            return self.hosts[dname]
        except KeyError:
            return defer.fail(Exception('Bad host passed to getMX'))

def makeSendmail(L):
    def sendmail(host, from_, toList, msgFile):
        L.append((host, from_, toList, msgFile))
        return defer.succeed(None)
    return sendmail

class Transmission(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [credup.StoreRealm, sched.SystemScheduler]

    def dbSetUp(self):
        self.checker = credup.ICredentialsChecker(self.store)
        self.avatar = self.createAvatar([smtpout.OutgoingMailup, EmailPowerup], "USAR")

        self.xmit = smtpout.IOutgoingMail(self.avatar)
        # self.xmit.mx.resolver.resolvers[2]._parseCall.cancel()

    def dbtestConvenience(self):
        self.xmit.getCalculator = lambda: MXHijacker({'thathost': defer.succeed(Record_MX(exchange="mxhost"))})
        L = []
        self.xmit.sendmail = makeSendmail(L)

        self.xmit.sendString("HELLO WORLD", [], 'thatguy@thathost', '', 'thisguy@thishost')

        self.assertEquals(len(L), 1)
        self.assertEquals(L[0][0], "mxhost")
        self.assertEquals(L[0][1], "thisguy@thishost")
        self.assertEquals(L[0][2], ["thatguy@thathost"])

    def dbtestFailedSendStaysInOutbox(self):
        # Jack the MX process so we can intercept the email
        d = defer.Deferred()
        self.xmit.getCalculator = lambda: MXHijacker({'thathost': d})

        L = []
        self.xmit.sendmail = makeSendmail(L)

        self.xmit.sendString("Message text", [], "thatguy@thathost", '', "thisguy@thishost")

        out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
        sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))
        self.assertEquals(len(out), 1)
        self.assertEquals(len(sent), 0)

        addrs = [('thatguy@thathost', 450, 'Transient forged failure')]
        exc = smtp.SMTPDeliveryError(-1, "One or more addresses not accepted", None, addrs)
        d.errback(failure.Failure(exc))

        out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
        sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))
        self.assertEquals(len(out), 1)
        self.assertEquals(len(sent), 0)

        d.result = None    # I don't owe you for World any more, JML -glyph

    def dbtestPermanentFailedGeneratesBounce(self):
        # Jack the MX process so we can intercept the email
        d = defer.Deferred()
        self.xmit.getCalculator = lambda: MXHijacker({'thathost': d})

        L = []
        self.xmit.sendmail = makeSendmail(L)

        self.xmit.sendString("Message text", [], "thatguy@thathost", '', "thisguy@thishost")
        out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
        sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

        self.assertEquals(len(out), 1)
        self.assertEquals(len(sent), 0)

        addrs = [('thatguy@thathost', 550, 'Permanent forged failure')]
        exc = smtp.SMTPDeliveryError(-1, "One or more addresses not accepted", None, addrs)
        d.errback(failure.Failure(exc))

        for i in range(25):
            reactor.iterate()

        out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
        sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

        self.assertEquals(out, [])
        self.assertEquals(sent, [])

        inbox = list(IMIMEDelivery(self.avatar).arrivalRef.getItem().queryIndex())
        self.assertEquals(len(inbox), 1)

        d.result = None    # I don't owe you for World any more, JML -glyph

class DeliveryException(Exception):
    pass

class FakeMIMEDelivery:
    __implements__ = (IMIMEDelivery,)

    def __init__(self, raiseWhen, count):
        self.raiseWhen = raiseWhen
        self.count = count

    def createMIMEReceiver(self, trustReceivedHeaders, x, y):
        return self

    def makeConnection(self, t):
        if self.raiseWhen == 'makeConnection':
            self.count -= 1
            if self.count <= 0:
                raise DeliveryException()

    def lineReceived(self, line):
        if self.raiseWhen == 'lineReceived':
            self.count -= 1
            if self.count <= 0:
                raise DeliveryException()

    def connectionLost(self, reason):
        if self.raiseWhen == 'connectionLost':
            self.count -= 1
            if self.count <= 0:
                raise DeliveryException()

    def messageDone(self):
        if self.raiseWhen == 'messageDone':
            self.count -= 1
            if self.count <= 0:
                raise DeliveryException()

class DeliveryWithErrors(unittest.TestCase, helper.ProvisionerMixin):
    avatarDomain = "localhost"

    def dbSetUp(self):
        self.avatar = self.createAvatar([EmailPowerup], "testavatar")
        self.portal = portal.Portal(credup.IRealm(self.store), [checkers.AllowAnonymousAccess(), credup.ICredentialsChecker(self.store)])
        self.factory = smtpin.ESMTPInFactory(self.portal)
        self.factory.startFactory()
        self.server = self.factory.buildProtocol(('localhost', 12345))
        self.transport = proto_helpers.StringTransport()
        self.server.makeConnection(self.transport)

    def dbTearDown(self):
        self.factory.stopFactory()
        # XXX Bug in StringTransport
        self.server.connectionLost(Exception("la la"))

    def dbtestConnectionMade(self):
        self.avatar.setComponent(IMIMEDelivery, FakeMIMEDelivery('makeConnection', 1))
        server, transport = self.server, self.transport

        server.lineReceived("HELO example.com")
        server.lineReceived("MAIL FROM: user@example.com")
        server.lineReceived("RCPT TO: testavatar@localhost")
        transport.clear()
        server.lineReceived("DATA")
        result = transport.value()
        self.assertEquals(len(result.splitlines()), 1)
        self.failUnless(result.startswith("550 "), "Got %r instead of a 550" % (result,))

        errors = log.flushErrors(DeliveryException)
        self.assertEquals(len(errors), 1)

    def _dbtestLineReceived(self, n):
        self.avatar.setComponent(IMIMEDelivery, FakeMIMEDelivery('lineReceived', n))
        server, transport = self.server, self.transport

        server.lineReceived("HELO example.com")
        server.lineReceived("MAIL FROM: user@example.com")
        server.lineReceived("RCPT TO: testavatar@localhost")
        transport.clear()
        server.lineReceived("DATA")
        result = transport.value()
        transport.clear()
        self.assertEquals(len(result.splitlines()), 1)
        self.failUnless(result.startswith("354 "), "Got %r instead of a 354" % (result,))

        server.lineReceived("Here is part of the message")
        result = transport.value()
        errors = log.flushErrors(DeliveryException)
        self.assertEquals(len(errors), 1)

    def dbtestLineReceivedOnlyBody(self):
        self._dbtestLineReceived(2)

    def dbtestLineReceived(self):
        self._dbtestLineReceived(3)

    def dbtestConnectionLost(self):
        self.avatar.setComponent(IMIMEDelivery, FakeMIMEDelivery('connectionLost', 1))
        server, transport = self.server, self.transport

        server.lineReceived("HELO example.com")
        server.lineReceived("MAIL FROM: user@example.com")
        server.lineReceived("RCPT TO: testavatar@localhost")
        transport.clear()
        server.lineReceived("DATA")
        server.lineReceived("here is message text rar rar rar")
        server.connectionLost(Exception("Reason"))

        errors = log.flushErrors(DeliveryException)
        self.assertEquals(len(errors), 1)


    def dbtestMessageDone(self):
        self.avatar.setComponent(IMIMEDelivery, FakeMIMEDelivery('messageDone', 1))
        server, transport = self.server, self.transport

        server.lineReceived("HELO example.com")
        server.lineReceived("MAIL FROM: user@example.com")
        server.lineReceived("RCPT TO: testavatar@localhost")
        server.lineReceived("DATA")
        server.lineReceived("Here is part of the message")
        transport.clear()
        server.lineReceived(".")
        result = transport.value()

        self.assertEquals(len(result.splitlines()), 1)
        self.failUnless(result.startswith("550 "), "Got %r instead of a 550" % (result,))

        errors = log.flushErrors(smtp.SMTPServerError, DeliveryException)
        self.assertEquals(len(errors), 2)

class Factories(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [sched.SystemScheduler]

    def dbSetUp(self):
        self.avatar = self.createAvatar([EmailPowerup, smtpout.OutgoingMailup], "av1")

    def dbTearDown(self):
        dc = sched.IScheduler(self.store).nextCall
        dc and dc.cancel()

    def dbtestInFactoryBuildProtocol(self):
        portal = object()
        f = smtpin.ESMTPInFactory(portal)
        f.doStart()
        p = f.buildProtocol(("remotehost", 12345))
        self.assertIdentical(p.factory, f)
        self.assertIdentical(p.deliveryFactory, f)
        self.assertIdentical(p.portal, portal)
