# -*- test-case-name: "quotient.test.test_webui" -*-

import os, re, sys
from os import path as osp
from os.path import join as opj

from twisted.python import log, components, hook
from twisted.spread import pb
from twisted.python.util import spewer, println

from quotient.functional import client, unittest
from quotient.functional import navtags
from quotient.functional.navtags import *

from nevow import guard

AuthData = unittest.AuthData

prefData =  {'realname': 'Mr. Tester',
             'replyto': 'test@example.com',
             'timezone': '2',
             'emergencyContact': 'otheraddress@example.net',
             'itemsPerPage':'10',
             'showInvisible':'1',
             'multipart': True}

grabberData = {'username': 'test',
               'password': 'pass',
               'password____2': 'pass',
               'host': 'example.com',
               'port': '995',
               'ssl': '1',
               'address': 'tester@example.com',
               'msgsPer':'50'}

priAndSecNav = {link('leftnav.Mail'): ['Compose', 'Mail Folders', 'Mailing Lists', 'Spam', 'Inbox'],
                link('leftnav.People'): ['Everybody', 'Address Book'],
                link('leftnav.Find'): ['Extracts', 'Explore', 'Search'],
                link('leftnav.Divmod'): ['Control Panel', 'Summary']}

# -- Bookmarks --

SETTINGS = [link('leftnav.Divmod'),
            link('SecNav.Control Panel')]

PERSONAL_PREFS = [SETTINGS,
                 link('Table.Personal Preferences')]

MAIL_SETTINGS = [link('leftnav.Mail'),
             link('SecNav.Mail Settings')]

NO_TRACEBACK = assertNotString('SorryRenderer')

LOGOUT = [url("/%s" % guard.LOGOUT_AVATAR), assertString('Claim a Ticket')]

metapup = url('/me/')


class LinkChecking(unittest.TestCase):
    login = []
    includeTelnetPBInTac = True
##     HTTPPort = 40080     # for easier post-run debugging
##     HTTPSPort = 40443
    __debug = False

    def setUpClass(self):
        if self.__debug:
            log.addObserver(self._debugObserver)

        self._dbgmsg('setUpClass')
        unittest.TestCase.setUpClass(self)
        
        from quotient.test import testdata
        from twisted.python import util
        self.asciiCorpusPath = util.sibpath(testdata.__file__, 'ascii')
        self.unicodeCorpusPath = util.sibpath(testdata.__file__, 'unicode')
        
        # use of genAuthAndSignup saves generated AuthData objects to this instance attribute
        self.authData = AuthData('tester3', 'sekrit', 'Mr. Test') # default

        self.tktSignup = {'desiredUsername': 'tktRedeemer',
                          'desiredPassword': 'asdf',
                          'desiredPassword____2': 'asdf'}

        self.tktSignin = {'username': self.tktSignup['desiredUsername'],
                          'password': self.tktSignup['desiredPassword']}

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.writeTestName()

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.runSeq(url('/'), LOGOUT)

    def _debugObserver(self, events):
        if self.__debug and 'linkChecking' in events:
            print ' '.join(['LinkChecking:\t'] + list(events['linkChecking']))

    def _dbgmsg(self, msg):
        log.msg(linkChecking=msg)

    def test_0000_noop(self):
        self.runSeq(url('/'))
        self._dbgmsg('if debug logging is enabled you should see this message')

    def test_0001_incorrectLogin(self):
        """test for incorrect login behavior

        Human Script:
          1. go to root page
          2. enter some garbage values in 'username' and 'password' fields and click 'go'
          3. make sure 'incorrect login' appears above 'username' field
        """
        self.runSeq(url('/'),
                    assertString('Divmod'),
                    form('signin', username='test', password='asdfawerqwerqwer'),
                    assertString('Incorrect login.'))

##     test_0001_incorrectLogin.todo = "Sometimes this times out, but that's no big deal"


    def test_0002_signUp(self):
        """sign up a new user

        Human Script:
          1. go to root page
          2. click 'Sign Up a New User' link
          3. Enter username, password, and real name
          4. make sure you are redirected to the 'Personal Preferences' page
        """
        self.authData = AuthData(u='test0003', p='sekrit')
        self.runSeq(self.doSignup,
                    assertString('Personal Preferences'))

    ## add test for signup as existing user error

    def test_0003_signin(self):
        """sign in after sign up

        Human Script:
          1. go to root page
          2. enter the username and password you used in L{test_0002_signUp}
          3. click 'go'
          4. make sure you don't see 'Incorrect login.'
          5. you should be redirected inside the app
            - exact page you are redirected to may change
        """
        
        self.runSeq(self.doSignin,
                    NO_TRACEBACK,
                    assertNotString('Incorrect login.'))

                    
    def test_0010_CheckPriSecNavs(self):
        """test each primary and secondary navigation link

        Human Script:
          1. sign in (if you aren't already)
          2. for each primary nav tab
            - make sure tab is highlighted when selected
            - click on each secondary nav link
              - make sure there are no tracebacks
              - make sure correct primary nav is highlighted
              - make sure secondary nav is rendered properly
        """
        self.authData = AuthData(u='test0010', p='sekrit')

        def genSeq():
            for pri, names in priAndSecNav.iteritems():
                yield tuple([pri] + [(link('SecNav.%s' % name), NO_TRACEBACK) for name in names])
            
        self.runSeq(self.doSignup, url('/me/'), NO_TRACEBACK, genSeq())


    def test_0015_PersonalPrefPost(self):
        """test personal preferences form

        Human Script:
          1. sign in
          2. go to Divmod > Control Panel > Personal Preferences
          3. enter appropriate data and change selections of drop-down boxes
          4. hit 'Submit Query'
          5. make sure there is no traceback
        """
        self.authData = AuthData(u='test0015', p='sekrit')
        self.runSeq(self.doSignup,
                    PERSONAL_PREFS,
                    lambda :url(client.page.url, **prefData),
                    assertString('Logout'),
                    NO_TRACEBACK)

    # TO WRITE:
    #    - check for blank textbox error reporting


    def doSimpleTableLinkTest(self, initialSeq=[], tableNameList=[]):
        """tests a table to make sure that all links can be reached"""
        self.runSeq(initialSeq)

        parentpage = client.page
        
        self.runSeq([[link('Table.%s' % name), NO_TRACEBACK, url(parentpage.url)] for name in tableNameList])


    def test_0040_Divmod(self):
        """test the links in the settings tab

        Human Script:
          1. sign in
          2. click on Divmod > Control Panel
          3. click on all links and make sure you don't get a traceback
        """
        s = ['Cheap Powerup Loading Trick',
             'Personal Preferences',
             'Resource Meter',
             'Scheduled Tasks',
             'Ticket Issuer',
             'Trash Man']

        self.authData = AuthData(u='test0040', p='sekrit')
        self.doSimpleTableLinkTest([self.doSignup, SETTINGS], s)


    def test_0043_Find_Extracts(self):
        """test extraction

        Human Script:
          1. sign in
          2. Find > extracts
          3. click on all links and make sure you don't get a traceback
        """
        s = ['Recipient Routing',
             'Gallery',
             'Attachments',
             'URLs',
             'Phone Numbers',
             'Questions',
             'Everybody',
             'Mailing Lists']
        self.authData = AuthData(u='test0043', p='sekrit')
        self.doSimpleTableLinkTest([self.doSignup, link('leftnav.Find'), link('SecNav.Extracts')], s)

    def test_0045_Explore(self):
        """test explore

        Human Script:
        1. sign in
        2. Find > explore
        3. click on all links and make sure you don't get a traceback
        """
        s = ['Calls',
             'Control Panel',
             'Extracts',
             'Mail Folders',
             'People',
             'Primary Navigation',
             'Trash']
        self.authData = AuthData(u='test0045', p='sekrit')
        self.doSimpleTableLinkTest([self.doSignup, link('leftnav.Find'), link('SecNav.Explore')], s)

    def test_0070_addPopGrabber(self):
        """add a pop grabber

        Human Script:
          1. sign in
          2. Mail > Mail Settings > External Accounts
          3. click 'Add Pop Grabber'
          4. fill in the form and submit
          5. make sure there is no traceback and that the pop grabber has been added
        """
        self.authData = AuthData(u='test0070', p='sekrit')
        self.runSeq(self.doSignup,
                    MAIL_SETTINGS,
                    link('Table.External Accounts'),
                    lambda :url(client.page.url, view='addPopGrabber'),
                    assertString('Add Pop Grabber'),
                    form('addPopGrabber', **grabberData),
                    NO_TRACEBACK,
                    assertString('POP3GrabberFactory'))


    def test_0080_addImapGrabber(self):
        """add an imap grabber

        Human Script:
          1. sign in
          2. Mail > Mail Settings > External Accounts
          3. click 'Add IMAP Grabber'
          4. fill in the form and submit
          5. make sure there is no traceback and that the imap grabber has been added
        """
        self.authData = AuthData(u='test0080', p='sekrit')
        self.runSeq(self.doSignup,
                    MAIL_SETTINGS,
                    link('Table.External Accounts'),
                    lambda :url(client.page.url, view = 'addIMAPGrabber'),
                    assertString('Add IMAP Grabber'),
                    form('addIMAPGrabber', **grabberData),
                    NO_TRACEBACK,
                    assertString('IMAP4GrabberFactory'))

    # TO WRITE:
    #     - grabber deletion
    #     - check time after 'check now' clicked


    def test_0120_TicketIssuer(self):
        """in trunk, test the ticket issuer. 

        Human Script:  (I{note, the protocol for doing this test on erin will be slightly different})
          1. sign in
          2. Settings > Ticket Issuer
          3. click 'issue ticket'
          4. make sure that a ticket is displayed with <username>-<#>-<random garbage> and copy that ticket number
          5. logout
          6. go to sign in page and click 'redeem a ticket'
          7. paste the ticket number and fill in information
          8. logout
          9. login with the new account information
        """
        self.authData = AuthData(u='test0120', p='sekrit')
        self.runSeq(self.doSignup,
                    SETTINGS,
                    link('Table.Ticket Issuer'),
                    form('issueTicket', value='submit'),
                    assertString(self.authData.signin['username']),
                    NO_TRACEBACK)

        m = re.search(r'(?P<ticket>%s-0-\w*\b)' % self.authData.signin['username'], client.page.domstr)
        self.failUnless(m, "Couldn't find ticket number on page %s" % client.page.domstr)
        log.msg(m.group('ticket'))
        self.tktSignup['ticket'] = m.group('ticket')

        self.runSeq(LOGOUT,
                    link('claim'),
                    form('redeemTicket', **self.tktSignup),
                    NO_TRACEBACK,
                    url('/'),
                    LOGOUT,
                    form('signin', **self.tktSignin),
                    NO_TRACEBACK)
        

    def test_0130_ComposePage(self):
        """test the compose page

        Human Script:
          1. sign in
          2. Mail > Compose
          3. fill in information in every field
          4. click 'send'
          5. make sure there is no traceback
          6. if appropriate, check that the email was sent to your indicated recipients
        """
        data = {'fromAddress': '0',
                'to': 'noone@example.net',
                'subj': 'this is a subject',
                'text': """this is the
text of a message

true not a very interesting message, but this is only a test.

if this were an actual message sent from quotient, you would be reading
some incredibly intelligent, witty, inciteful text here.


but it isn't, so it isn't



why are you still reading this?



piss off!



no, seriously, piss off!


Sincerely,
Richard M. Nixon
""",
                #'text': 'a simple message',
                'cc': 'rmnixon@example.net',
                'bcc': 'deepthroat@example.com',
                'replyTo': 'woodard_and_bernstein@example.com',
                'multipart': True}
        
        self.authData = AuthData(u='test0130', p='sekrit')
        self.runSeq(self.doSignup,
                    link('leftnav.Mail'),
                    link('SecNav.Compose'),
                    form('sendMail', **data),
                    NO_TRACEBACK,
                    assertString("'Send Mail'"),
                    assertString("successfully"))

    test_0130_ComposePage.todo = "I don't know why buildbot hates this test, it works locally"

    # TO WRITE:
    #     - create a second user and send mail to them, log out and log in as
    #       second user and make sure delivery works
    #
    #     - test error on blank fields and no error with blank message body
    
    def test_0140_trashMan(self):
        """test trash man

        Human Script:
          1. sign in
          2. Settings > Trash Man
          3. change the age of the collection threshold and submit
          4. click empty all
        """
        self.authData = AuthData(u='test0140', p='sekrit')
        self.runSeq(self.doSignup,
                    SETTINGS,
                    link('Table.Trash Man'),
                    assertString('Collection Threshold'),
                    lambda :url(client.page.url, age='10'),
                    NO_TRACEBACK,
                    SETTINGS,
                    link('Table.Trash Man'),
                    form('emptyAll', submit='Submit'),
                    NO_TRACEBACK)

    def test_0300_SearchTests_loadMailAndIndex(self):
        # make a new user for this test
        ad = self.authData = AuthData(u='test4', p='sekrit')
        user = ad.signin['username']
        
        self.runSeq(self.doSignup, self.doSignin, assertNotString('Incorrect login'))  # XXX: blech! fix0rate!+

        sekritHQ = self.sekritHQ

        sekritHQ.makeAdmin(user)
        sekritHQ.loadMailDir(user, self.asciiCorpusPath)
        r = sekritHQ.indexingTestSettings(user)
        self.assert_(r, 'allowIndexing returned %s' % r)
        self._dbgmsg('waiting for indexing')
        sekritHQ.waitForIndexing(user)

        storeID = self.sekritHQ.getAvatarStoreID(user)
        self._dbgmsg('storeID = %s' % storeID)
        sekritHQ.forceIndexerFlush(user)

        indexerDir = opj(self.tdir, self.deployConfig['fileDirectory'], str(storeID), 'indexer')

        self._dbgmsg('indexerDir = %s' % indexerDir)

        self.assert_(len(os.listdir(indexerDir)) > 1, "Indexer did _not_ flush its files")

        query = 'dazzle'

        r = sekritHQ.indexerSanityCheck(user, query)
        self.assert_(r, 'indexer is insane') 

        searchupStoreID = sekritHQ.getSearchupStoreID(user)
        
        self.runSeq(self.doSignin,
                    url('/me/%d' % searchupStoreID, query=query, _charset_="UTF-8"),
                    assertNotString('No matches found'), 
                    assertString(query))


    def test_0310_SearchTests_crazyComplete(self):
        self.runSeq(self.doSignin)
        user = self.authData.signin['username']

        wordlist = ['addition', 'addressees', 'blouse', 'blurt', 'concatenations',
                    'conceits', 'debark', 'deans', 'emphatic', 'emulates',
                    'fingerprinting', 'financier']
        
        def _testWordlist():
            searchupStoreID = self.sekritHQ.getSearchupStoreID(user)
            for word in wordlist:
                yield (url('/me/%d' % searchupStoreID, query=word, _charset_="UTF-8"),
                        assertNotString('No matches found'), 
                        assertString(word))

        self.runSeq(self.doSignin, _testWordlist())

    def test_regression_ZeroLengthUsername(self):
        """test to make sure that not entering a username in the signup page causes an error message to be displayed"""
        self.runSeq(url('/'),
                    link('signup'),
                    form('signup',
                         username="",
                         password="sekrit",
                         password____2="sekrit",
                         realname="blanker"),
                    assertString('Please enter a string'))

##     test_regression_ZeroLengthUsername.todo = "Don't know why this is failing"


    def test_regression_Issue305(self):
        # http://divmod.org:8000/rt/Ticket/Display.html?id=305
        ad = AuthData(u='test305', p='sekrit')                # don't save this AuthData as we don't need it later
        self.runSeq(lambda :self.doSignup(**ad.signup))       # i'm not sure why this is necessary to do in runSeq...
        self.runPbRootCommand('loadMailDir', ad.signin['username'], self.asciiCorpusPath)

    test_regression_Issue305.skip = 'skip this right now'


    def test_divmod_tab(self):
        """tests the unread/new counts in the divmod tab
        1. import a directory full of mail
        2. log in
        3. for each pool 'foo' in the list
           - click on the 'foo' link
           - make sure that all 'new' messages are rendered in the tdb (i.e. page through all
             pages where there are 'new' messages)
           - click the divmod tab
           - make sure that the pool 'foo' has been removed from the list in the divmod tab
        4. import more mail (it's okay if it's duplicates)
        5. for each pool 'foo' in the list
           - make sure that the 'unread' count is higher than the 'new' count
        """


    msg = "skipped for testing"
##     test_0001_incorrectLogin.skip = msg
##     test_0002_signUp.skip = msg    
##     test_0003_signin.skip = msg
##     test_0010_CheckPriSecNavs.skip = msg
##     test_0015_PersonalPrefPost.skip = msg
##     test_0040_Divmod.skip = msg
##     test_0043_Find_Extracts.skip = msg
##     test_0045_Explore.skip = msg
##     test_0070_addPopGrabber.skip = msg
##     test_0080_addImapGrabber.skip = msg
##     test_0120_TicketIssuer.skip = msg
##     test_0130_ComposePage.skip = msg
##     test_0140_trashMan.skip = msg
##     test_0150_whitelist.skip = msg
##     test_0300_SearchTests_loadMailAndIndex.skip = msg
##     test_0310_SearchTests_crazyComplete.skip = msg
##     test_regression_ZeroLengthUsername.skip = msg
##     test_regression_Issue305.skip = msg

