# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.
import time, sys

from twisted.internet import defer, reactor
from twisted.python import log
from twisted.python.components import Interface, CannotAdapt
from twisted.application.service import IService, IServiceCollection

from atop import store
from atop import powerup

from atop.store import Item, Pool, registerIndex, referenced, indexed
from atop.powerup import Storeup
from atop.service import Serviceup

from quotient.web import root

class ILoggedTraceback(Interface):
    """a traceback that was logged by quotient"""

class LoggedTraceback(Item):
    __implements__ = ILoggedTraceback,

    text = ''

    def __init__(self, store,
                 text='The Traceback',
                 tm=None):
        Item.__init__(self, store, name='Traceback')
        self.text = text
        self.summary = text.splitlines()[-2]

        if not tm:
            tm = time.time()
        self.time = tm

    def __str__(self):
        return str(self.text)


class ITraceup(Interface):
    """For when the excrement strikes the rotating blades"""

class Traceup(Storeup):
    __implements__ = (Storeup.__implements__, ITraceup, IService)

    storeDependencies = [Serviceup]
    powerupName = "Traceback Recorder"

    tbPool = store.referenced
    running = False

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        s = powerup.IPowerStation(self.store).installPowerup(Serviceup)
        self.name = self.powerupName
        s.addService(self)

    def setUpPools(self, store):
        store.setComponent(ITraceup, self)
        self.tbPool = Pool(self.store, 'TracebackPool')
        IServiceCollection(self.store).addService(self)

    def privilegedStartService(self):
        raise AssertionError("privilegedStartService *cannot* be called!")

    def startService(self):
        self.touch()
        self.running = True
        log.addObserver(self.observe)

    def stopService(self):
        self.touch()
        self.running = False
        log.removeObserver(self.observe)

    def addMsg(self, msg, time=None):
        self.tbPool.addItem(LoggedTraceback(self.store, msg, time))

    def observe(self, events):
        if events['message']:
            return
        if events['isError'] and 'failure' in events:
            text = events['failure'].getTraceback()
            fmtStr = "%Y/%m/%d %H:%M:%S %Z"
            logmsg = time.strftime(fmtStr, time.localtime(events['time']))
            logmsg = logmsg + " [%s] %s\n" % (events['system'], text.replace("\n", "\n\t"))
            self.store.transact(self.addMsg, logmsg)

    def callWithoutObserving(self, f, *a, **kw):
        """Run a function without observing its log output.
        """
        self.stopService()
        try:
            return f(*a, **kw)
        finally:
            self.startService()

    def allYourTbs(self):
        return self.tbPool.queryIndex()

    def getPastTracebacks(self, timedelta):
        """get tracebacks for timedelta from now"""
        pass

    def expireTracebacksOlderThan(self, timedelta):
        """get rid of tracebacks older than now - timedelta"""
        pass

class TabularLoggedTraceback(root.TabularBase):
    def render_summary(self, ctx, data):
        return data.summary
    root.column(render_summary, True)

    def render_time(self, context, data):
        # stolen from voiceup.py
        # convert to datetime object
        local = tpython.mkgmtime(data.time)
        dtobj = datetime.datetime.fromtimestamp(local, timezones.Local)
        # and here's where we have to respect the user's account setting, hoorj
        astz = dtobj.astimezone(prefup.IPrefs(data.store).timezone)
        return astz.strftime('%Y/%m/%d %H:%M:%S')
    root.column(render_time, True)
